import json
import logging
import time

import requests

try:
    import urllib.parse as urlparse
except ImportError:
    import urlparse


class CommunicationObject(object):
    def __init__(
        self,
        url,
        project_id,
        token,
        timeout=900.0,
        chunk_size=256,
        verify=True,
        critical=False,
    ):
        """
        Communications with the platform

        Parameters
        ----------
        url : str
            Platform URL
        project_id : int
            Unique project identifier
        token : str
            Unique token that authenticates the communication
        timeout : float or tuple, optional
            How many seconds to wait for the server to send data before giving up, as a float, or a
            (connect timeout, read timeout) tuple. Defaults to
        chunk_size : int
            Size of chunks in bytes (1MB = 1024KB, 1KB = 1024B)
        verify : bool, optional
            Controls whether the server's TLS certificate is verified.
            Defaults to True.
        critical: bool, optional
            Overrides total_retry seting in requests by making themm fail after the first error if set.
        """
        self.timeout = timeout
        self.chunk_size = chunk_size

        self.__req_url = url
        self.__project_id = project_id
        self.__token = token
        self.__verify = verify
        self.__critical = critical

        self.__session = requests.Session()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        ret = self.send_request("logout")
        if not ret["success"]:
            # Not being able to logout is a security issue and must be a fatal error
            raise RuntimeError("Unable to logout")

    def send_request(
        self,
        access_point=None,
        req_data=None,
        req_headers=None,
        return_json=True,
        timeout=None,
        total_retry=5,
        stream=False,
        url=None,
        method="POST",
        **kwargs
    ):
        """
        Send a request to the platform to the given API endpoint.

        Parameters
        ----------
        access_point : str, optional
            API endpoint. Will be joined with the request URL provided at construction time to create the
            full ("absolute") url
        req_data : dict or str, optional
            Request data
        req_headers : dict, optional
            Request headers
        return_json : bool, optional
            Whether the response should be a JSON object or a requests.Response instance.
            Defaults to True.
        timeout : float, optional
            Timeout in seconds for this request.
            Defaults to the timeout set at construction time.
        total_retry : int, optional
            Number of connection retries.
            Defaults to 5.
        stream : bool, optional
            If False, the response content will be immediately downloaded.
        url : str, optional
            When no access_point is provided, a full url must be provided through this argument
        kwargs : dict
            Additional arguments to be passed to `requests.Session.post`
        method: str
            HTTP verb (for instance POST or GET)

        Returns
        -------
        requests.Response or dict
            JSON response encoded in a dictionary if return_json is True (default),
            otherwise requests.Response object.
        """
        logger = logging.getLogger(__name__)

        # Default parameters
        req_data = req_data or dict()
        req_headers = req_headers or dict()
        timeout = timeout or self.timeout

        if access_point:
            # Add authentication for the platform
            req_headers["Cookie"] = "AUTH_COOKIE={}; MINT_AP={}".format(
                self.__token, self.__project_id
            )
            req_headers["Mint-Api-Call"] = "1"
            url = urlparse.urljoin(self.__req_url, access_point)

        retry_counter = 0
        while True:
            try:
                response = self.__session.request(
                    method=method,
                    url=url,
                    data=req_data,
                    headers=req_headers,
                    timeout=timeout,
                    verify=self.__verify,
                    stream=stream,
                    **kwargs
                )

                # log & raise error if we get a HTTP error code
                if not response.ok:
                    logger.debug(
                        "{!r} HTTP response: {!r}".format(url, response.text)
                    )
                response.raise_for_status()

            except (
                requests.HTTPError,
                requests.Timeout,
                requests.ConnectionError,
            ):
                retry_counter += 1
                if self.__critical or total_retry == retry_counter:
                    raise
                time.sleep(retry_counter * 5)

            else:
                return json.loads(response.text) if return_json else response

    def send_files(
        self,
        access_point,
        files,
        req_data=None,
        req_headers=None,
        return_json=True,
        timeout=None,
        total_retry=5,
        stream=False,
        **kwargs
    ):
        """
        Send a POST request to the platform to the given API endpoint.

        Parameters
        ----------
        access_point : str
            API endpoint. Will be joined with the request URL provided at construction time to create the
            full ("absolute") url
        files : dict
            Request files
        req_data : dict or str, optional
            Request data
        req_headers : dict, optional
            Request headers
        return_json : bool, optional
            Whether the response should be a JSON object or a requests.Response instance.
            Defaults to True.
        timeout : float, optional
            Timeout in seconds for this request.
            Defaults to the timeout set at construction time.
        total_retry : int, optional
            Number of connection retries.
            Defaults to 5.
        stream : bool, optional
            If False, the response content will be immediately downloaded.
        kwargs : dict
            Additional arguments to be passed to `requests.Session.post`

        Returns
        -------
        requests.Response or dict
            JSON response encoded in a dictionary if return_json is True (default),
            otherwise requests.Response object.
        """
        logger = logging.getLogger(__name__)

        # Default parameters
        req_data = req_data or dict()
        req_headers = req_headers or dict()
        timeout = timeout or self.timeout

        # Add authentication for the platform
        req_headers["Cookie"] = "AUTH_COOKIE={}; MINT_AP={}".format(
            self.__token, self.__project_id
        )
        req_headers["Mint-Api-Call"] = "1"

        retry_counter = 0
        while True:
            try:
                url = urlparse.urljoin(self.__req_url, access_point)
                response = self.__session.post(
                    url,
                    files=files,
                    data=req_data,
                    headers=req_headers,
                    timeout=timeout,
                    verify=self.__verify,
                    stream=stream,
                    **kwargs
                )

                # Raise error if we get a HTTP error code
                response.raise_for_status()
                result = json.loads(response.text) if return_json else response
                logger.info("{!r} HTTP response: {!r}".format(url, result))
                return result

            except (
                requests.HTTPError,
                requests.Timeout,
                requests.ConnectionError,
            ):
                retry_counter += 1
                if self.__critical or total_retry == retry_counter:
                    raise
                time.sleep(retry_counter * 5)


class DummyCommunicationObject(object):
    def __init__(self, url, chunk_size=1, mock_response=None):
        self.__req_url = url
        self.chunk_size = chunk_size
        self.mock_response = mock_response
        self.sent_contents = None

    def send_request(
        self,
        access_point=None,
        req_data=None,
        req_headers=None,
        return_json=True,
        timeout=None,
        total_retry=5,
        stream=False,
        url=None,
        method="POST",
        **kwargs
    ):
        self.sent_contents = req_data
        if stream and method == "GET":
            return DummmyResponse()
        elif self.mock_response is None or stream:
            return DummmyResponse()
        else:
            return self.mock_response


class DummmySession(object):

    def get(self, url, stream=False):
        if stream:
            return DummmyResponse()
        else:
            return "{}"


class DummmyResponse(object):

    def iter_content(self, *args, **kwargs):
        return []



import hashlib
import json
import logging
import os
import re
import time
import uuid
import zipfile
from base64 import b64encode
from enum import Enum

from requests import HTTPError

from qmenta.sdk.directory_utils import TemporaryDirectory


class NoFilesError(Exception):
    """Error raised when no files are found in a `data` analysis type."""

    pass


class AnalysisContext:
    class QCEntity(Enum):
        """Enum with the following options: ANALYSIS, SESSION."""

        ANALYSIS = "analysis"
        SESSION = "patients"

    class QCStatus(Enum):
        """Enum with the following options: FAIL, PASS."""

        PASS = "pass"
        FAIL = "fail"

    class ReplaceFlag(Enum):
        """Enum with the following options: ALL, ONLY_FILE."""

        NO_REPLACEMENT = "0"
        ONLY_FILE = "1"
        ALL = "2"

    def __init__(self, analysis_id, comm):
        """Context object that interfaces with the QMENTA platform.
        Should not be created directly.
        """
        self.analysis_id = analysis_id
        self.__comm = comm
        self.__containers_data = {}
        self.analysis_data = None
        self.parent_analysis_data = None
        self.__initial_retry_wait_time = 0.01  # seconds
        self.__max_upload_retries = 5
        self.logger = logging.getLogger(__name__)
        self.__initial_chunk_retry_wait_time = (
            0.25  # seconds, TODO: remove when URL upload is fixed
        )
        self.__max_chunk_retries = 8

    def fetch_analysis_data(self):
        """Fetch information about current analysis.

        Returns
        -------
        dict
            Dictionary that contains important information about the current
            analysis, such as:
            * state (str) : State of the analysis
            * name (str) : Name of the analysis
            * script_name (str) : Name of the script
            * analysis_id (str) : A unique identifier for the analysis
            * user_id (str) : Username of the user that started analysis
            * patient_secret_name (str) : Name of the subject
            * ssid (str) : Timepoint identifier
            * settings (dict) : Settings dictionary.
            * tags (list) : List of tags
            * version (str) : Tool version
        """
        res = self.__comm.send_request(
            "analysis_manager/get_analysis_list", {"id": self.analysis_id}
        )
        if res:
            self.analysis_data = res[0]
            self.analysis_data["analysis_id"] = str(self.analysis_id)
            return self.analysis_data
        else:
            error_msg = "Analysis with ID={} not found".format(
                self.analysis_id
            )
            raise ValueError(error_msg)

    def fetch_parent_analysis_data(self):
        """Fetch information about current analysis.

        Returns
        -------
        dict
            Dictionary that contains important information about the current
            analysis, such as:
            * state (str) : State of the analysis
            * name (str) : Name of the analysis
            * script_name (str) : Name of the script
            * analysis_id (str) : A unique identifier for the analysis
            * user_id (str) : Username of the user that started analysis
            * patient_secret_name (str) : Name of the subject
            * ssid (str) : Timepoint identifier
            * settings (dict) : Settings dictionary.
            * tags (list) : List of tags
            * version (str) : Tool version
        """
        if self.parent_analysis_data is not None:
            return self.parent_analysis_data

        if not self.has_parent_analysis():
            raise Exception(
                "The current analysis does not have parent analysis !"
            )

        parent_analysis_id = self.analysis_data["super_analysis"][
            "parent_analysis_id"
        ]

        res = self.__comm.send_request(
            "analysis_manager/get_analysis_list", {"id": parent_analysis_id}
        )
        if res:
            self.parent_analysis_data = res[0]

            # fetching the full state of a flow
            res_state = self.__comm.send_request(
                "analysis_manager/get_super_analysis_state",
                {"analysis_id": parent_analysis_id},
            )

            self.parent_analysis_data["nodes"] = (
                res_state.get("data", dict()) if res_state else dict()
            )

            return self.parent_analysis_data
        else:
            error_msg = "Analysis flow with ID={} not found".format(
                parent_analysis_id
            )
            raise ValueError(error_msg)

    def get_communication_object(self):
        """Return the communication object used for communication with
        platform.

        Returns
        -------
        CommunicationObject
            The communication object used to communicate the platform
        """
        return self.__comm

    def set_qc_status(
        self,
        status=QCStatus.PASS,
        comments="",
        entity=QCEntity.ANALYSIS,
        input_id=None,
    ):
        """Changes the analysis or session QC status.

        Parameters
        ----------
        status : QCStatus
            QCStatus.PASS or QCStatus.FAIL
        comments : str, optional
            Additional comments explaining why the QC status has been set to
            pass or fail.
        entity : QCEntity, optional
            QCEntity.ANALYSIS or QCEntity.SESSION
        input_id : str, optional
            The id of the session in the settings definition when entity is
            QCEntity.SESSION
        """
        if entity == self.QCEntity.ANALYSIS:
            item_id = self.analysis_id
        elif entity == self.QCEntity.SESSION:
            container_id = self.analysis_data["settings"][input_id][
                "container_id"
            ]
            res = self.__comm.send_request(
                "file_manager/get_container_files",
                {"container_id": container_id},
            )
            item_id = str(int(res["patient_id"]))
        else:
            raise RuntimeError("Unknown entity {}".format(entity))

        data_to_send = {
            "item_ids": item_id,
            "status": status.value,
            "comments": comments,
            "entity": entity.value,
        }

        self.logger.info(
            "Setting ({}={}) QC status to {}: {}".format(
                entity.name, item_id, status, comments
            )
        )

        res = self.__comm.send_request(
            "projectset_manager/set_qa_status", data_to_send
        )
        if not res["success"]:
            raise RuntimeError(res["error"])

    def get_qc_status(self, analysis=False):
        """Gets the session QC status.

        Parameters
        ----------
        analysis: bool
            to get analysis QC
        """
        qa_status, qa_comments = None, None
        if not analysis:
            data_sessions = self.__comm.send_request(
                "patient_manager/get_patient_list",
                req_headers={"X-Range": "items=0-9999"},
                req_data={
                    "pars_patient_secret_name": "string"
                    ";{}".format(self.analysis_data['patient_secret_name'])
                },
            )
            self.logger.info(data_sessions[0])
            self.logger.info(
                "Getting QC for {patient_secret_name} / {ssid}".format(
                    **self.analysis_data
                )
            )
            for session in data_sessions:
                ssid = session["ssid"]
                self.logger.info(
                    "Matches QC for {} / {}?".format(
                        self.analysis_data["patient_secret_name"], ssid
                    )
                )

                if str(self.analysis_data["ssid"]) == str(ssid):
                    self.logger.info(session)
                    qa_status = session["qa_status"]
                    qa_comments = session["qa_comments"]
                    break

        if analysis:
            data_analyses = self.__comm.send_request(
                "analysis_manager/get_analysis_list",
                req_headers={"X-Range": "items=0-9999"},
                req_data={
                    "secret_name": self.analysis_data["patient_secret_name"]
                },
            )
            try:
                for analysis in data_analyses:
                    if analysis["_id"] == self.analysis_id:
                        self.logger.info(
                            f"Found analysis / {analysis}, getting QC."
                        )
                        qa_data = analysis["qa_data"]["new_value"]
                        qa_status = qa_data.split("<b>")[1].split("</b>")[
                            0
                        ]  # `fail` or `pass`
                        qa_comments = qa_data.split("<br/>")[1]
                        break
            except Exception as e:
                # Handle other potential exceptions
                self.logger.info(f"An error occurred: {e}")
        return qa_status, qa_comments

    def download_session_qa_requirements(self, path):
        """Downloads the list of project-level requirements for a session.

        Parameters
        ----------
        path : str
            The location where to store the requirements.r
        """
        if os.path.exists(path):
            raise RuntimeError(
                "Cannot download session qa because {} already exists!".format(
                    path
                )
            )

        data_to_send = {"project_id": self.analysis_data["projectset_id"]}

        url = "projectset_manager/get_session_qa_requirements"
        res = self.__comm.send_request(url, data_to_send)
        if not res["success"]:
            raise RuntimeError("Cannot download the session QA requirements")
        else:
            self.logger.info(
                "Storing project-level session QA requirements at {}".format(
                    path
                )
            )
            with open(path, "w") as fs:
                # Process as JSON to validate
                json.dump(res["data"], fs)

    def upload_session_qa_report(self, input_id, path):
        """Sets the QA status for all the files of session as per requirements.

        Parameters
        ----------
        input_id : str
            Identifier "id" of the input container in advanced settings.
        path : str
            The json file with the QA status for all files and rules
        """
        with open(path, "r") as fp:
            # Load as JSON to verify structure
            data = json.loads(fp.read())

        container_id = self.analysis_data["settings"][input_id]["container_id"]
        data_to_send = {"container_id": container_id, "data": json.dumps(data)}
        self.logger.info(
            "Sending session QA report with id {}:\n{}".format(
                data_to_send["container_id"], data_to_send["data"]
            )
        )

        res = self.__comm.send_request(
            "file_manager/set_session_qa", data_to_send
        )
        if not res["success"]:
            raise RuntimeError(res["error"])

    def set_progress(self, message=None, value=None):
        """Sets analysis progress.

        Parameters
        ----------
        value : int, optional
            Number between 0 and 100 indicating the current status of execution
        message : str, optional
            Progress message.
        """
        if not value and not message:
            self.logger.warning(
                "No information is available to report the progress"
            )
            return

        data_to_send = {"analysis_id": self.analysis_id}

        if value:
            value = int(value)  # Cast to int to avoid float/string problems
            if value > 100:
                self.logger.warning(
                    "The progress value should be between 0 and 100 "
                    "(using 100 instead)"
                )
                value = 100
            elif value < 0:
                self.logger.warning(
                    "The progress value should be between 0 and 100 "
                    "(using 0 instead)"
                )
                value = 0
            data_to_send["value"] = value

        if message:
            data_to_send["message"] = message

        self.logger.info("Progress: {} -> {}".format(value, message))
        self.__comm.send_request(
            "analysis_manager/set_analysis_progress", data_to_send
        )

    def get_settings(self):
        """Analysis settings.

        Returns
        -------
        dict
            Settings for the current analysis. This includes all the parameters
            defined in the tool specification (checkboxes, input fields, etc.)
            and can be accessed using their identifier.
        """
        return self.analysis_data["settings"]

    def get_files(
        self,
        input_id,
        modality=None,
        tags=None,
        reg_expression=None,
        file_filter_condition_name=None,
        subject_container_id=None,
    ):
        """Returns files in input container specified by `input_id` that match.

        Returns files that match the modality, tags and regular expression, if
        any. (X)OR that are selected by the file filter condition defined in
        the tool settings.

        Parameters
        ----------
        input_id : str
            Identifier "id" of the input container in advanced settings.
        modality : str, optional
            Optional modality, will allow any modalities if None given.
            Default is None.
        tags : set, optional
            Optional set of tags, will allow any tags if None given, and will
            match any set of tags (including the empty set) otherwise.
            Default is None.
        reg_expression : str, optional
            Regular expression to match with the file names in the specified
            container, will allow any file names if None given.
            Default is None.
        file_filter_condition_name : str, optional
            File filter specified in the settings of the tool.
            Default is None.
        subject_container_id : str, optional

        Examples
        --------
        Get all files that passed c_TEST (including any selection the user may
        have done using the GUI)

        >>> context.get_files('input', file_filter_condition_name='c_TEST')

        Get all files that passed c_TEST and have no tags

        >>> context.get_files('input', file_fitler_condition_name='c_TEST',
        ...                   tags=set())

        Get all files that passed c_TEST and have the DTI tag in their list of
        tags

        >>> context.get_files('input', file_fitler_condition_name='c_TEST',
        ...                   tags={'DTI'})

        Get all files that passed c_TEST whose name match the given regexp

        >>> context.get_files('input', file_filter_condition_name='c_TEST',
        ...                  reg_expression='^some_expression[0-9]')

        Alternatively, you can use no file_filter_condition_name to browse
        through the whole input container

        >>> context.get_files('input')

        Returns
        -------
        list of qmenta.sdk.context.File
            List of selected file handlers from the container.
        """
        self.logger.info(
            "Get files: {} [modality={} / tags={} / regex={}]".format(
                input_id, modality, tags, reg_expression
            )
        )

        file_filter_condition_name = file_filter_condition_name or ""

        input_settings = self.analysis_data["settings"][input_id]
        is_subject_setting = "containers" in input_settings
        container_id = (
            subject_container_id
            if is_subject_setting
            else input_settings["container_id"]
        )

        if is_subject_setting and subject_container_id is None:
            msg = (
                "must define subject_container_id for subjects "
                "with multiple sessions"
            )
            self.logger.error(msg)
            raise Exception(msg)

        if is_subject_setting:
            # get input settings for correct container
            try:
                all_input_setting = [
                    r
                    for r in input_settings["containers"]
                    if r["container_id"] == container_id
                ]
                input_settings = all_input_setting[0]
            except KeyError:
                msg = "container {} not found for input {!r}".format(
                    container_id, input_id
                )
                self.logger.exception(msg)
                raise Exception(msg)

            # ensure data structure is compatible with expected data
            if input_id not in self.__containers_data:
                self.__containers_data[input_id] = {}

            have_container_data = (
                container_id in self.__containers_data[input_id]
            )

        else:
            have_container_data = input_id in self.__containers_data

        if not have_container_data:
            # If files from the file container are not available, fetch them
            res = self.__comm.send_request(
                "file_manager/get_container_files",
                {"container_id": container_id},
            )

            if is_subject_setting:
                self.__containers_data[input_id][container_id] = res["data"]
            else:
                self.__containers_data[input_id] = res["data"]

        container_data_to_work_with = self.__containers_data[input_id]
        if is_subject_setting:
            container_data_to_work_with = self.__containers_data[input_id][
                container_id
            ]

        if file_filter_condition_name:
            # Get the files that passed the file_filter and take those that
            # have the required conditions
            selected_files = []

            if input_settings["passed"]:
                try:
                    file_filter_spec = input_settings["filters"][
                        file_filter_condition_name
                    ]
                except KeyError:
                    msg = "File filter condition {} not specified in the \
                    container {} settings !".format(
                        file_filter_condition_name, input_id
                    )
                    self.logger.error(msg)
                    raise Exception(msg)

                for file_desc in file_filter_spec["files"]:
                    metadata = container_data_to_work_with["meta"]
                    find_file_data = [
                        f for f in metadata if f["name"] == file_desc["name"]
                    ]
                    if find_file_data:
                        file_data = find_file_data[0]

                        fname, fmetadata, ftags = (
                            file_data["name"],
                            file_data["metadata"],
                            set(file_data["tags"]),
                        )

                        if _should_include_file(
                            fname,
                            fmetadata.get("modality", None),
                            ftags,
                            modality,
                            tags,
                            reg_expression,
                        ):
                            selected_files.append(
                                File(
                                    container_id,
                                    fname,
                                    fmetadata,
                                    ftags,
                                    self.__comm,
                                )
                            )

            return selected_files
        else:
            # in case the file filter condition name is not specified then we
            # need to take the other concrete search conditions i.e.
            # modality, tags, reg_expression

            # filter the files according to the filter defined
            return self.__files_from_container_data(
                container_data_to_work_with,
                container_id,
                modality,
                tags,
                reg_expression,
            )

    def _get_files_from_container(
        self, container_id, modality=None, tags=None, reg_expression=None
    ):
        self.logger.info("Get files from container {}".format(container_id))
        res = self.__comm.send_request(
            "file_manager/get_container_files", {"container_id": container_id}
        )
        if res["success"]:
            return self.__files_from_container_data(
                res["data"], container_id, modality, tags, reg_expression
            )
        else:
            raise RuntimeError(res["error"])

    def __files_from_container_data(
        self, container_data, container_id, modality, tags, reg_expression
    ):
        selected_files = []
        for file_data in container_data["meta"]:

            # TODO: Here we are assuming that file metadata has a name and a
            # metadata attribute. Need of jsonschema

            fname, fmetadata, ftags = (
                file_data["name"],
                file_data["metadata"],
                set(file_data["tags"]),
            )
            assume_file_selected = _should_include_file(
                fname,
                fmetadata.get("modality", None),
                ftags,
                modality,
                tags,
                reg_expression,
            )

            if assume_file_selected:
                selected_files.append(
                    File(container_id, fname, fmetadata, ftags, self.__comm)
                )
        return selected_files

    @staticmethod
    def __detect_file_format(file_path):
        if file_path.endswith(".nii"):
            return "nifti"
        elif file_path.endswith(".nii.gz"):
            return "nifti"
        elif file_path.endswith(".bvec"):
            return "bvec"
        elif file_path.endswith(".bval"):
            return "bval"
        else:
            return None

    def new_scan_session(
        self,
        ssid=None,
        date_at_scan=None,
        age_at_scan=None,
        backup_age=None,
        patient_metadata_id=None,
    ):
        """Add a new scan session.

        Only functional for analysis tools of type `data`.

        Parameters
        ----------
        ssid : int, optional
        date_at_scan : str, optional
            Possible formats: '%d.%m.%Y' / '%Y-%m-%d'.
            Examples: '01.01.2000' / '2000-01-01'
        age_at_scan : int, optional
        backup_age : int, optional
        patient_metadata_id : int, optional

        Returns
        -------
        int
            ID of the new scan session
        """
        try:
            patient_secret_name = self.analysis_data["patient_secret_name"]
        except KeyError:
            raise RuntimeError(
                "Patient secret name not available in analysis information. "
                "Check platform failures."
            )
        try:
            md5_original_file = self.get_settings()["md5_original_file"]
        except KeyError:
            raise RuntimeError(
                "MD5 hash not available via analysis settings. "
                "Check platform failures."
            )

        data = {
            "patient_secret_name": patient_secret_name,
            "md5_original_file": md5_original_file,
            "date_at_scan": date_at_scan,
            "age_at_scan": age_at_scan,
            "backup_age": backup_age,
            "analysis_id": self.analysis_id,
        }

        if ssid is not None:
            data["ssid"] = ssid

        if patient_metadata_id:
            data["patient_metadata_id"] = patient_metadata_id

        reply = self.__comm.send_request(
            "patient_manager/add_scan_session", data
        )

        if reply["success"]:
            return reply["data"]
        else:
            raise Exception(reply["error"])

    @staticmethod
    def __valid_upload_response(response, accumulated_bytes, total_bytes):
        """Files are sent chunk by chunk.

        Backend returns first a string per chunk and a JSON when the upload
        is completed (last chunk)
        Example:
          Response chunk 0:     0-524287/10485760 (str)
          Response chunk 1:     0-1048575/10485760 (str)
          ...
          Response chunk N-1:   0-9961471/10485760 (str)
          Response chunk N:     {"success": 1} (json)
        TODO: remove when upload via URL is working again
        """
        logger = logging.getLogger(__name__)

        try:
            if accumulated_bytes == total_bytes:  # Final chunk
                if response.json()["success"]:
                    return True
                else:
                    raise RuntimeError("Final chunk upload failed")
            else:  # Intermediate chunk
                expected_response = "0-{}/{}".format(
                    accumulated_bytes - 1, total_bytes
                )
                if response.text == expected_response:
                    return True
                else:
                    raise RuntimeError(
                        "Intermediate chunk upload \
                          failed: {!r} != {!r}".format(
                            response.text, expected_response
                        )
                    )
        except Exception as e:
            logger.error(
                "Error when loading upload response: {!r} -> {!r}".format(
                    response.text, e
                )
            )
            return False

    @staticmethod
    def __file_md5(path):
        BLOCKSIZE = 65536
        hasher = hashlib.md5()
        with open(path, "rb") as fp:
            buf = fp.read(BLOCKSIZE)
            while len(buf) > 0:
                hasher.update(buf)
                buf = fp.read(BLOCKSIZE)
        return b64encode(hasher.digest())

    def __validate_upload_response(self, response, total_bytes, md5_hash):
        if response.status_code != 200:
            self.logger.error(
                "Upload returned status code {}".format(response.status_code)
            )
            return False

        upload_manifest = response.json()
        if int(upload_manifest["size"]) != int(total_bytes):
            self.logger.error(
                "Upload size mismatch: {} != {}".format(
                    upload_manifest["size"], total_bytes
                )
            )
            return False

        if upload_manifest["md5Hash"] != md5_hash:
            self.logger.error(
                "MD5 hash mismatch: {} != {}".format(
                    upload_manifest["md5Hash"], md5_hash
                )
            )
            return False

        return True

    def upload_file(
        self,
        source_file_path,
        destination_path,
        modality=None,
        tags=None,
        file_info=None,
        file_format=None,
        protocol=None,
        container_id=None,
        replace=None,
        direct=True,
    ):
        """Upload a file to the platform, to be part of the result files.

        Parameters
        ----------
        source_file_path : str
            Path to the file to be uploaded.
        destination_path : str
            Path in the output container. Will be shown in the platform.
        modality: str, optional
            Optional modality of the uploaded file.
            None by default.
        tags : set, optional
            Set of tags for the uploaded file.
            None by default.
        file_info : dict, optional
            File information metadata.
        file_format : str, optional
            Use "dicom" when .zip file with .dcm slices (uploading a "nifti"
            file is automatically recognized as such).
        protocol : str, optional
        container_id : str, optional
        replace : str, optional
        direct : bool
        """
        tags = tags or set()
        # File information
        source_file_path = str(source_file_path)
        destination_path = str(destination_path)
        file_name = destination_path.split("/")[-1]
        total_bytes = os.path.getsize(source_file_path)

        self.logger.info(
            "Requesting upload URL to upload {!r} to {!r}".format(
                source_file_path, destination_path
            )
        )
        if not total_bytes > 0:
            self.logger.warning(
                "Skipping empty file {} upload".format(source_file_path)
            )
            return

        if modality:
            tags.add("m:" + modality)

        if replace:
            assert isinstance(replace, AnalysisContext.ReplaceFlag)
        else:
            replace = (
                AnalysisContext.ReplaceFlag.NO_REPLACEMENT
            )  # Try to find a proper file format when not available

        if not direct:
            session_id = str(uuid.uuid4())
            # make direct upload to bucket until URL is fixed
            chunk_num = 0
            with open(source_file_path, "rb") as file_object:
                chunk_retry_count = 0
                wait_time = self.__initial_chunk_retry_wait_time
                data = file_object.read(self.__comm.chunk_size)
                # If chunk num is 0, we still want to send an empty file
                while (
                    data or chunk_num == 0
                ) and chunk_retry_count <= self.__max_chunk_retries:
                    start_position = chunk_num * self.__comm.chunk_size
                    self.logger.debug(
                        "Chunk size is {}".format(self.__comm.chunk_size)
                    )
                    end_position = start_position + self.__comm.chunk_size - 1
                    bytes_to_send = self.__comm.chunk_size

                    if end_position >= total_bytes:
                        end_position = total_bytes - 1
                        bytes_to_send = (
                            total_bytes - chunk_num * self.__comm.chunk_size
                        )

                    bytes_range = "bytes {}-{}/{}".format(
                        start_position, end_position, total_bytes
                    )

                    # Request headers
                    req_headers = {
                        "Content-Type": "application/octet-stream",
                        "Content-Range": bytes_range,
                        "Session-ID": str(session_id),
                        "Content-Length": str(bytes_to_send),
                        "Content-Disposition": "attachment; "
                        "filename={}".format(file_name),
                    }

                    # If it is the last chunk, define more header fields
                    accumulated_bytes = (
                        chunk_num * self.__comm.chunk_size + bytes_to_send
                    )
                    if accumulated_bytes == total_bytes:
                        self.logger.debug("Sending last data chunk")
                        req_headers.update(
                            {
                                "X-Mint-Analysis-Output": str(
                                    self.analysis_id
                                ),
                                "X-Mint-File-Destination-Pat\
                                    h": destination_path,
                                "X-Mint-File-Tags": ",".join(tags),
                            }
                        )

                        # Conditional headers
                        if file_info:
                            req_headers["X-Mint-File-Info"] = json.dumps(
                                file_info
                            )
                        if file_format:
                            req_headers["X-Mint-File-Format"] = str(
                                file_format
                            )
                        if protocol:
                            req_headers["X-Mint-Protocol"] = str(protocol)
                        if container_id:
                            req_headers["X-Mint-Container"] = str(container_id)
                        if replace:
                            assert isinstance(
                                replace, AnalysisContext.ReplaceFlag
                            )
                            req_headers["X-Mint-Replace-Flag"] = replace.value
                        req_headers["X-Requested-With"] = "XMLHttpRequest"

                    # Send data chunk
                    response = self.__comm.send_request(
                        "upload", data, req_headers, return_json=False
                    )

                    # Verify upload
                    if self.__valid_upload_response(
                        response, accumulated_bytes, total_bytes
                    ):
                        # Read next chunk
                        data = file_object.read(self.__comm.chunk_size)
                        chunk_num += 1
                        chunk_retry_count = 0
                        wait_time = self.__initial_chunk_retry_wait_time
                    else:
                        chunk_retry_count += 1
                        wait_time = 2 * wait_time
                        time.sleep(wait_time)

                if (
                    data or chunk_retry_count != 0
                ):  # Restarted when successful upload
                    msg = (
                        "File {} could not be uploaded to the platform.\n"
                        "data - {}\n"
                        "chunk_retry_count - {}".format(
                            source_file_path, data, chunk_retry_count
                        )
                    )
                    self.logger.error(msg)
                    raise RuntimeError(msg)
                else:
                    self.logger.info("Upload completed")

        else:
            request_data = {
                "type": "analysis_output",
                "size": total_bytes,
                "path": destination_path,
                "analysis_id": self.analysis_id,
                "replace_flag": replace.value,
                "tags": ",".join(tags),
                "file_info": json.dumps(file_info),
                "file_format": file_format or self.__detect_file_format(source_file_path),
                "protocol": str(
                    protocol or file_name.split(".")[0]
                    if "." in file_name
                    else file_name
                ),
                "container_id": container_id,
            }
            url_request_response = self.__comm.send_request(
                "file_manager/get_upload_url", req_data=request_data
            )

            if (
                not url_request_response["success"]
                or "url" not in url_request_response
            ):
                self.logger.error("Error when requesting upload URL")
                self.logger.error("Request data: {}".format(request_data))
                self.logger.error(
                    "Returned error: {}".format(
                        url_request_response.get("error", None)
                    )
                )
                raise RuntimeError("Failed Upload")

            signed_url = url_request_response["url"]
            container_id = url_request_response["container_id"]

            md5_hash = self.__file_md5(source_file_path).decode()

            # Stream file to signed-url without loading it to memory
            completed_upload = False
            retry_count = 0
            wait_time = self.__initial_retry_wait_time
            while (
                not completed_upload
                and retry_count < self.__max_upload_retries
            ):
                with open(source_file_path, "rb") as fp:
                    upload_response = self.__comm.send_request(
                        method="POST",
                        url=signed_url,
                        req_data=fp,
                        return_json=False,
                    )

                # Verify upload
                if self.__validate_upload_response(
                    upload_response, total_bytes, md5_hash
                ):
                    completed_upload = True
                else:
                    retry_count += 1
                    if retry_count < self.__max_upload_retries:
                        wait_time = 2 * wait_time
                        time.sleep(wait_time)
                        self.logger.warning(
                            "Retrying upload (#{}) in {}".format(
                                retry_count, wait_time
                            )
                        )

            if not completed_upload:
                msg = "File {} could not be uploaded to the platform".format(
                    source_file_path
                )
                self.logger.error(msg)
                raise RuntimeError(msg)

            verification_response = self.__comm.send_request(
                "file_manager/verify_upload_via_url",
                req_data={
                    "container_id": container_id,
                    "path": destination_path,
                    "md5": md5_hash,
                },
            )
            if not verification_response.get("success", False):
                msg = "The upload could not be verified: {}".format(
                    verification_response.get("error")
                )
                self.logger.error(msg)
                raise RuntimeError(msg)
            else:
                self.logger.info("Upload verified")

    def download_resource(self, resource_path, destination_path):
        """Downloads a resource file from the bucket.

        Downloads user/group based resources, or legacy.

        Parameters
        ----------
        resource_path : str
            Path to the file in the resources bucket.
        destination_path : str
            Path where the file will be downloaded.
        """
        self.logger.info(
            "Downloading resource {!r} to {!r}".format(
                resource_path, destination_path
            )
        )

        data_to_send = {
            "analysis_id": self.analysis_id,
            "type": type,
            "path": resource_path,
        }
        try:
            res = self.__comm.send_request(
                "analysis_manager/get_resource",
                data_to_send,
                total_retry=1,
                return_json=False,
                stream=True,
            )
        except HTTPError as ex:
            self.logger.exception(ex)
            if ex.response.status_code == 404:
                msg = "Resource file {!r} not found".format(resource_path)
                self.logger.error(msg)
                raise Exception(msg)

            raise

        os.makedirs(os.path.dirname(destination_path), exist_ok=True)

        self.logger.debug("Saving {!r}".format(destination_path))
        with open(destination_path, "wb") as fp:
            for chunk in res.iter_content(chunk_size=self.__comm.chunk_size):
                if chunk:  # filter out keep-alive new chunks
                    fp.write(chunk)
                    fp.flush()
        self.logger.debug(
            "Finished downloading resource {!r}".format(destination_path)
        )

    def _get_manual_analysis_data(self):
        """Returns manual analysis data generated during manual step.

        Returns
        -------
        dict
            The values generated during the manual step (user interaction).
        """
        self.logger.info("Getting manual analysis data")

        data_to_send = {"analysis_id": self.analysis_id}
        try:
            res = self.__comm.send_request(
                "analysis_manager/get_manual_analysis_data",
                data_to_send,
                total_retry=1,
            )
        except HTTPError as ex:
            self.logger.exception(ex)
            if ex.response.status_code == 404:
                msg = "Manual analysis data not found (request exception)"
                self.logger.error(msg)
                raise Exception(msg)
            raise
        else:
            self.logger.info(res)
            return res["value"]

    def set_metadata_value(
        self,
        key,
        value,
        title=None,
        readonly=False,
        patient_secret_name=None,
        ssid=None,
    ):
        """Sets the value of a metadata parameter.

        Parameters
        ----------
        key : str
            The ID of the metadata parameter.
        value : int, str, float or list
            The new content of the parameter.
        title : str, optional
            How the metadata field should be presented
        readonly : bool, optional
            Whether the user should be able to edit the value on the platform
            or not (only by analyses)
        patient_secret_name : str, optional
            Manually select the patient_secret_name
        ssid : str, optional
            Manually select the ssid. This is useful to choose a specific
            session in a longitudinal analysis.
        """
        if not isinstance(key, str):
            raise Exception("Parameter 'key' must be of type 'str'.")
        result = re.findall(r"[^a-zA-Z0-9_\s]", key)
        if result:
            raise Exception(
                f"Special characters not accepted: {', '.join(result)}"
            )
        if not isinstance(value, (int, list, str, float)):
            raise Exception(
                "Parameter 'value' must be of type 'int', 'list', 'str', "
                "or 'float'."
            )

        if not self.get_metadata_value(key):
            # Create the metadata field if it does not exist
            if title:
                assert isinstance(title, str)

            allowed_types = {
                int: "integer",
                list: "list",
                str: "string",
                float: "decimal",
            }

            if type(value) not in allowed_types:
                raise ValueError(
                    "Invalid type for the metadata value: {!r}".format(
                        type(value)
                    )
                )
            parameter_type = allowed_types[type(value)]
            self._assure_metadata_parameters(
                [
                    {
                        "key": key,
                        "title": title if title else key,
                        "type": parameter_type,
                        "readonly": int(readonly),
                    }
                ]
            )

        self.logger.info(
            "Setting metadata parameter: {} = {!r}".format(key, value)
        )

        data_to_send = {
            "analysis_id": self.analysis_id,
            "key": key,
            "value": value,
            "patient_secret_name": patient_secret_name
            or self.analysis_data["patient_secret_name"],
            "ssid": ssid or self.analysis_data["ssid"],
        }

        try:
            res = self.__comm.send_request(
                "analysis_manager/set_metadata_value",
                data_to_send,
                total_retry=1,
            )
            if not res["success"]:
                msg = "Metadata error: {!r}".format(res["error"])
                self.logger.error(msg)
                raise Exception(msg)
        except HTTPError as ex:
            self.logger.exception(ex)
            raise

    def get_metadata_value(
        self, key=None, patient_secret_name=None, ssid=None
    ):
        """Gets the value of a metadata parameter for the current session.

        Parameters
        ----------
        key : str, optional
            The ID of the metadata parameter. If not set, it will return all
            the metadata for the session.
        patient_secret_name: str, optional
            Manually select the patient_secret_name
        ssid : str, optional
            Manually select the ssid. This is useful to choose a specific
            session in a longitudinal analysis.

        Returns
        -------
        The value of the parameter if it exists, raises otherwise
        """
        self.logger.info(
            "Getting value from field with id {} from subject {}/{}".format(
                key, patient_secret_name, ssid
            )
        )

        data_to_send = {
            "analysis_id": self.analysis_id,
            "ssid": ssid or self.analysis_data["ssid"],
            "patient_secret_name": patient_secret_name
            or self.analysis_data["patient_secret_name"],
        }

        if key is not None:
            data_to_send["key"] = key

        try:
            res = self.__comm.send_request(
                "analysis_manager/get_metadata_value",
                data_to_send,
                total_retry=1,
            )
            if not res["success"]:
                if "error" in res and "no metadata field" in res["error"]:
                    raise ValueError("Metadata field is empty or not found.")
                else:
                    msg = "Metadata error: {!r}".format(res["error"])
                    self.logger.error(msg)
                    raise RuntimeError(msg)
        except HTTPError as ex:
            self.logger.exception(ex)
            raise
        else:
            if res.get("value", None) is not None:
                return res["value"]
            else:
                raise ValueError("Metadata field is empty or not found.")

    def _assure_metadata_parameters(self, params):
        """Ensure that a parameter exists.

        Parameters
        ----------
        params : list
            The definition of the parameter. It may contain "title", "type",
            "visible", "readonly" or "order".
        """
        if not isinstance(params, list):
            params = list(params)

        for data_to_send in params:
            if not isinstance(data_to_send, dict):
                msg = (
                    "Format error. Metadata param configuration should be "
                    "stored in a dictionary"
                )
                self.logger.error(msg)
                raise TypeError(msg)

            data_to_send["analysis_id"] = self.analysis_id
            self.logger.info(
                "Checking metadata param config: {}".format(data_to_send)
            )
            try:
                res = self.__comm.send_request(
                    "analysis_manager/set_metadata_param",
                    data_to_send,
                    total_retry=1,
                )
                if not res["success"]:
                    msg = "Metadata error: {!r}".format(res["error"])
                    self.logger.error(msg)
                    raise Exception(msg)
            except HTTPError as ex:
                self.logger.exception(ex)
                raise

    # For multi-timepoint (longitudinal) analyses
    def get_sessions(self, input_id):
        input_data = self.analysis_data["settings"][input_id]
        try:
            containers = input_data["containers"]
        except KeyError:
            msg = "settings for {!r} must be from a subject"
            self.logger.error(msg)
            raise Exception(msg)
        else:
            return [
                Container(self, input_id, container_data)
                for container_data in containers
            ]

    def has_parent_analysis(self):
        return "super_analysis" in self.analysis_data

    def get_parent_files_from_input_id(self, input_id):
        if not self.has_parent_analysis():
            raise Exception(
                "The current analysis does not have parent analysis !"
            )

        filters_settings = (
            self.fetch_parent_analysis_data()
            .get("settings", dict())
            .get(input_id, dict())
            .get("filters", dict())
            .items()
        )
        selected_files = set()
        for filter_name, filter_content in filters_settings:
            if not filter_content["passed"]:
                continue
            for ff in filter_content.get("files", []):
                selected_files.add(ff["name"])

        return list(selected_files)

    def set_analysis_output_variables(
        self, variables_dict, clear_previous=False
    ):
        """Sets the output variables of analysis.

        Parameters
        ----------
        variables_dict : str
            Dictionary with the output variables
        clear_previous : bool, optional
            Whether the previous output variables should be cleared or not
        """
        if not isinstance(variables_dict, dict):
            self.logger.error(
                "The output variables are not of type dictionary ! "
                "(actual type {})".format(str(type(variables_dict)))
            )
            raise RuntimeError(
                "The output variables are not of type dictionary !"
            )

        variables_to_send = json.dumps(variables_dict)

        try:
            data_to_send = {
                "output": variables_to_send,
                "clear": int(clear_previous),
                "analysis_id": self.analysis_id,
            }

            res = self.__comm.send_request(
                "analysis_manager/set_output_variables",
                data_to_send,
                total_retry=2,
            )
            if not res["success"]:
                msg = "Setting output variables: {!r}".format(res["error"])
                self.logger.error(msg)
                raise Exception(msg)
        except HTTPError as ex:
            self.logger.exception(ex)
            raise

    def get_session_owner(self, input_id):
        """Get session owner.

        Parameters
        ----------
        input_id: str
            ID of the container used as input in the analysis

        Returns
        -------
        Owner information
        """
        self.logger.info(f"Getting value for owner of container {input_id}.")
        return self.__comm.send_request(
            "file_manager/get_container_files",
            self.analysis_data["settings"][input_id],
        )["owner"]


class Container:
    # is a kind of session, no other kinds of session currently implemented

    def __init__(self, context, input_id, container_data):
        self.__container_id = container_data.get("container_id", None)
        self.__input_id = input_id
        self.__context = context

        self.container_data = container_data

    def get_metadata(self):
        return self.__context._get_metadata_value(None)

    def get_files(
        self,
        modality=None,
        tags=None,
        reg_expression=None,
        file_filter_condition_name=None,
    ):
        return self.__context.get_files(
            self.__input_id,
            modality,
            tags,
            reg_expression,
            file_filter_condition_name,
            self.__container_id,
        )

    def _set_metadata_value(self, key, value):
        return self.__context._set_metadata_value(
            key,
            value,
            self.container_data["subject_name"],
            self.container_data["ssid"],
        )


class File:
    def __init__(self, container_id, name, metadata, tags, comm):
        """Object that represents a file and all its metadata in the platform.

        Should not be created directly.
        """
        self.__comm = comm
        self.__container_id = container_id
        self.name = name
        self.__metadata = metadata
        self.__tags = tags
        self.__download_path = None
        self.logger = logging.getLogger(__name__)

    def get_file_modality(self):
        """Get the modality of the file.

        Returns
        -------
        str or None
            Modality of the file or None if not known.
        """
        return self.__metadata.get("modality", None)

    def get_file_format(self):
        """Get the format of the file (e.g. 'nifti', 'txt', etc.).

        Returns
        -------
        str
            File format or None if not known.
        """
        return self.__metadata.get("format", None)

    def get_file_info(self):
        """Get the file information.

        The type of information depends on the type of file (e.g. nifti files
        include information such as 'Data strides', 'Data type', 'Dimensions'
        or 'Voxel size').

        Returns
        -------
        dict
            Dictionary of key-value pairs that depend on the format of file.
        """
        return self.__metadata.get("info", {})

    def get_file_tags(self):
        """Get the file tags.

        Returns
        -------
        set
            Set of tags associated to this file
        """
        return self.__tags if self.__tags is not None else set()

    def get_file_path(self):
        """Get the file download path.

        Returns
        -------
        str
            The path of the file where it has been downloaded

        Raises
        ------
        RuntimeError
            If `get_file_path` is called before the file has been downloaded
        """
        if self.__download_path is None:
            err_msg = (
                "File {!r} has no path because it has not been "
                "downloaded".format(self.name)
            )
            self.logger.error(err_msg)
            raise RuntimeError(err_msg)
        else:
            return self.__download_path

    def __download_file(self, destination_file_path):
        """Downloads the file into the specified `destination_file_path`.

        Parameters
        ----------
        destination_file_path : str
            Path where the file should be downloaded.
        """
        source_file = self.name

        self.logger.info(
            "Downloading {!r} to {!r}".format(
                source_file, destination_file_path
            )
        )

        data_to_send = {
            "container_id": self.__container_id,
            "files": source_file,
        }
        self.logger.debug(
            "Accessing {files!r} from container {container_id!r}".format(
                **data_to_send
            )
        )
        url_response = self.__comm.send_request(
            access_point="file_manager/download_file_via_url",
            req_data=data_to_send,
            return_json=True,
            stream=False,
        )

        if "data" not in url_response or len(url_response["data"]) != 1:
            raise RuntimeError(
                "Cannot download file via URL: {error_message}".format(
                    error_message=url_response.get("error", "No error message")
                )
            )
        presigned_entry = url_response["data"][0]  # Only one file
        if not isinstance(presigned_entry, dict):
            self.logger.info("Presigned URL {!r}".format(presigned_entry))
            presigned_entry = json.loads(presigned_entry)

        url = presigned_entry["url"]
        download_response = self.__comm.send_request(
            url=url, return_json=False, stream=True, method="GET"
        )

        os.makedirs(os.path.dirname(destination_file_path), exist_ok=True)
        with open(destination_file_path, "wb") as fp:
            # For Google Cloud Bucket, performance is at its best at 1 MB or
            # larger chunks.
            for chunk in download_response.iter_content(
                chunk_size=self.__comm.chunk_size
            ):  # Use 10 MB chunk
                if chunk:  # filter out keep-alive new chunks
                    fp.write(chunk)
                    fp.flush()  # Avoid using too much RAM by flushing to disk

        self.__download_path = destination_file_path
        self.logger.debug(
            "Finished downloading {!r}".format(destination_file_path)
        )

    def download(self, dest_path, unpack=True):
        """Downloads a file or the contents of a packed file.

        Parameters
        ----------
        dest_path:
            Path where the file should be downloaded to.
        unpack:
            Tells the function whether the file should be unpacked to the
            given folder.

        Returns
        -------
        str:
            The full path of the file or the folder with the unpacked files.
        """
        source_file = self.name

        # Normalize path
        dest_path = os.path.abspath(dest_path)

        self.logger.debug("Using path {!r}".format(dest_path))

        if source_file.endswith(".zip") and unpack:
            # Download the zip to a temporary directory and unpack its contents
            # to the user dir path
            with TemporaryDirectory() as temp_dir:
                temp_path = os.path.join(temp_dir, source_file)
                self.__download_file(temp_path)
                with zipfile.ZipFile(temp_path, "r") as zip_ref:
                    os.makedirs(dest_path, exist_ok=True)
                    zip_ref.extractall(path=dest_path)

            self.__download_path = (
                dest_path  # Replace download path with directory
            )
            return dest_path

        # Download the file to the directory and return the full path
        else:
            destination_file_path = os.path.join(dest_path, source_file)
            self.__download_file(destination_file_path)
            return destination_file_path


""" HELPERS """


def _should_include_file(
    fname,
    fmodality,
    ftags,
    search_modality,
    search_tags,
    search_reg_expression,
):
    """Checks if file should be selected or not according to specified
    selectors."""
    assert fname

    if search_modality is not None and fmodality != search_modality:
        return False

    # Search with tags has three outcomes:
    # 1. None - No filter (take all)
    # 2. Empty set - Take only those with no tags
    # 3. Set with elements - Take those with AT LEAST those tags
    if search_tags is not None:
        assert isinstance(ftags, set)
        assert isinstance(search_tags, set)
        if search_tags == set() and ftags != set():
            return False
        elif not search_tags.issubset(ftags):
            return False

    # If regexp, filter with it
    if search_reg_expression is not None:
        if not re.match(search_reg_expression, fname):
            return False

    return True



import argparse
import json
import logging
import os
import sys

import humanfriendly

from qmenta.sdk import __version__
from qmenta.sdk.client import ExecClient
from qmenta.sdk.communication import CommunicationObject
from qmenta.sdk.context import AnalysisContext

log_levels = {"info": logging.INFO, "debug": logging.DEBUG}

default_chunk_size = "10 MB"


def setup_logging(log_level, log_path=os.path.abspath("exec.log")):
    # Logging settings
    root_logger = logging.getLogger()
    formatter = logging.Formatter(
        "[%(asctime)s] %(levelname)s %(name)s:%(lineno)d :: %(message)s"
    )
    root_logger.setLevel(log_level)

    # File handler
    file_handler = logging.FileHandler(log_path)
    file_handler.setFormatter(formatter)
    root_logger.addHandler(file_handler)

    # Console handler
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(formatter)
    root_logger.addHandler(console_handler)

    # get urlib3 to stop spamming the logs
    logging.getLogger("requests.packages.urllib3.connectionpool").setLevel(
        logging.WARN
    )

    return log_path


def parse_args():
    metadata = (
        json.dumps({"version": __version__}, sort_keys=True, indent=4) + "\n"
    )

    argp = argparse.ArgumentParser(
        description="Executor for external developers (SDK version {})".format(
            __version__
        )
    )
    argp.add_argument("--project", type=int, required=True, help="Project ID")
    argp.add_argument(
        "--analysis", type=int, required=True, help="Analysis ID"
    )
    argp.add_argument("--url", required=True, help="Platform URL")
    argp.add_argument(
        "--token", required=True, help="Access token for the API"
    )
    argp.add_argument(
        "--tool-path",
        required=True,
        help="Path to the Python tool (package.module:function) to be executed",
    )

    # verbosity
    argp.add_argument(
        "--log-level", choices=log_levels, default="info", help="Logging level"
    )

    # chunk size of downloads / uploads
    argp.add_argument(
        "--chunk-size",
        type=str,
        default=default_chunk_size,
        help="Size of each chunk in download/upload operations, expressed in a human readable data size. "
        "(e.g. 256 --> 256 bytes ; 512B --> 512 bytes ; 1MiB --> 1048576 bytes; 1MB --> 1000000",
    )

    # only for testing purposes
    argp.add_argument(
        "--allow-insecure",
        type=bool,
        nargs="?",
        const=True,
        default=False,
        help="Avoid checking certificates for HTTPS requests",
    )
    argp.add_argument(
        "--no-retry",
        type=bool,
        nargs="?",
        const=True,
        default=False,
        help="Ignore requests total_retry number and fail after each first errror",
    )

    # metadata
    argp.add_argument("--version", action="version", version=__version__)
    argp.add_argument("--metadata", action="version", version=metadata)
    return argp.parse_args()


def main():
    arguments = parse_args()
    log_path = setup_logging(log_levels[arguments.log_level])

    # Prepare client to run analysis
    chunk_size = humanfriendly.parse_size(arguments.chunk_size, binary=True)
    with CommunicationObject(
        arguments.url,
        arguments.project,
        arguments.token,
        verify=not arguments.allow_insecure,
        chunk_size=chunk_size,
        critical=arguments.no_retry,
    ) as comm:

        context = AnalysisContext(arguments.analysis, comm)
        client = ExecClient(arguments.analysis, comm, context, log_path)

        logger = logging.getLogger(__name__)
        logger.info("Using SDK version: {}".format(__version__))

        # Run it
        try:
            client(arguments.tool_path)
        except Exception as e:
            logger.exception(e)
            raise SystemExit(1)


if __name__ == "__main__":
    main()



from __future__ import print_function
import os
import sys
from argparse import ArgumentParser

entry_point_contents = """#!/bin/bash -eu

export PYTHONPATH="${{PYTHONPATH:+$PYTHONPATH:}}{tool_path}"

# Add your configuration here:
# ...

# Tool start:
exec {executable} -m qmenta.sdk.executor "$@"
"""


def main():
    parser = ArgumentParser()
    parser.add_argument("target")
    parser.add_argument("tool_paths", nargs="+")
    parser.add_argument(
        "--example-tool-path",
        help="the directory should be included in tool_paths",
    )
    parser.add_argument("--example-tool-type", choices=tool_types)

    options = parser.parse_args()

    make_entrypoint(
        options.target,
        options.tool_paths,
        options.example_tool_path,
        options.example_tool_type,
    )


# Default tools
matlab_default_tool = """

import os

def run(context):
    # Get information from the platform such as patient name, user_id, ssid...
    analysis_data = context.fetch_analysis_data()
    # example_string = 'Analysis name: ' + analysis_data['name'] + '\n'

    # Get the complete list of input file handlers matching the requirements
    input_files = context.get_files('input', modality='T1')

    # Download all those files
    for file_handler in input_files:
        file_handler.download('/root/')  # Keeping the original names

        # Call the MATLAB helper script with the path of the MATLAB runtime installation
        # Make sure your script finalizes. Otherwise this call will be waiting forever
        # Alternatively, you can always implement a timeout (e.g. with subprocess)

        # os.system('/root/MATLAB_run_script.sh /opt/mcr/v92')

        # Last, upload the results to the platform
        # context.upload_file('/root/results.png', 'results.png')
"""

minimal_default_tool = """
def run(context):
    # Get information from the platform such as patient name, user_id, ssid...
    analysis_data = context.fetch_analysis_data()

    # Get the complete list of input file handlers
    input_files = context.get_files('input')  # You can add fitlers (modality, tags...)

    # Download them
    for file_handler in input_files:
        file_handler.download('/root/')

        # Do some processing here
        # ...

        # And upload the results
        # context.upload_file(path, path_in_output)  # You can add filters (modality, tags...)
"""

mrinfo_mrtrix_default_tool = """
import os

def run(context):
    # Get information from the platform such as patient name, user_id, ssid...
    analysis_data = context.fetch_analysis_data()

    with open('/root/sizes.txt', 'w') as fp:
        fp.write('Analysis name: ' + analysis_data['name'] + '\n')

    # Get the complete list of files
    input_files = context.get_files('input', modality='T1')

    # Do some processing (run mrinfo for every file)
    for f in input_files:
        f.download('/root/')
        os.system('mrinfo {} >> /root/sizes.txt'.format(f.get_file_path()))

    # Last, upload the results to the platform
    context.upload_file('/root/sizes.txt', 'mrinfo_outputs.txt')
"""

tool_types = {
    "minimal": minimal_default_tool,
    "matlab": matlab_default_tool,
    "mrtrix": mrinfo_mrtrix_default_tool,
}


def make_entrypoint(
    target,
    tool_paths,
    example_tool_path,
    example_tool_type,
    help_stream=sys.stderr,
):
    if not example_tool_path.endswith(".py"):
        raise RuntimeError(
            "expected tool path to be a python file to be created"
        )

    with open(target, "w") as fp:
        fp.write(
            entry_point_contents.format(
                executable=sys.executable, tool_path=":".join(tool_paths)
            )
        )

    if example_tool_path:
        with open(example_tool_path, "w") as fp:
            fp.write(tool_types.get(example_tool_type, "minimal"))

        print("execute:", file=help_stream)
        print(
            "$",
            target,
            os.path.basename(example_tool_path)[: -len(".py")] + ":run",
            file=help_stream,
        )
        print("to run the example tool", file=help_stream)


if __name__ == "__main__":
    main()



#! /usr/bin/env python

import json
import os
import re
from tkinter import filedialog, CENTER, END, LEFT, RIGHT, X, Y, BOTH, YES, HORIZONTAL, TOP, messagebox

import ttkbootstrap as ttk
from PIL import Image, ImageTk
from qmenta.core import platform
from qmenta.core.auth import Needs2FAError
from qmenta.sdk.tool_maker.make_files import raise_if_false

FONT = "Consolas"
FONTSIZE = 10

MIN = 0
MAX_CORES = 10
MAX_RAM = 16

dir_name = None


def gui_tkinter():
    def browse_folder():
        global dir_name
        dir_name = filedialog.askdirectory()
        write_dir.set(dir_name)
        tool_id_entry.delete(0, END)
        tool_id_entry.insert(0, os.path.basename(dir_name))

        # checking version
        version_file = os.path.join(dir_name, "version")
        with open(version_file) as fv:
            version_ = fv.read()
        tool_version_entry.delete(0, END)
        tool_version_entry.insert(0, version_)
        image_name_entry.delete(0, END)
        image_name_entry.insert(0, os.path.basename(dir_name) + f":{version_}")

    def generate_output_object():
        values = {
            "qmenta_user": user_id_entry.get(),
            "qmenta_password": password_entry.get(),
            "code": tool_id_entry.get(),
            "version": tool_version_entry.get(),
            "name": tool_name_entry.get(),
            "cores": num_cores_entry.get(),
            "memory": memory_entry.get(),
            "image_name": image_name_entry.get(),
            "docker_url": docker_url_entry.get(),
            "docker_user": docker_user_entry.get(),
            "docker_password": docker_password_entry.get(),
        }
        return values

    def perform_selection_check(values):
        try:
            raise_if_false(
                isinstance(values["code"], str), "Tool ID must be a string."
            )
            raise_if_false(values["code"] != "", "Tool ID must be defined.")
            raise_if_false(
                " " not in values["code"], "Tool ID can't have spaces."
            )
            values["code"] = values["code"].lower()  # must be lowercase
            values["short_name"] = (
                values["code"].lower().replace(" ", "_")
            )  # must be lowercase

            raise_if_false(
                isinstance(values["name"], str), "Tool name must be a string."
            )
            raise_if_false(values["name"] != "", "Tool name must be defined.")

            raise_if_false(
                values["version"] != "", "Tool version must be defined."
            )
            raise_if_false(
                re.search(r"^(\d+\.)?(\d+\.)?(\*|\d+)$", values["version"]),
                "Version format not valid.",
            )

            raise_if_false(values["cores"], "Number of cores must be defined.")
            raise_if_false(
                values["cores"].isnumeric(),
                "Number of cores must be an integer.",
            )
            raise_if_false(
                MAX_CORES >= int(values["cores"]) > MIN,
                f"Number of cores must be between {MIN} and {MAX_CORES}.",
            )

            raise_if_false(values["memory"], "RAM must be defined.")
            raise_if_false(
                values["memory"].isnumeric(), "RAM must be an integer."
            )
            raise_if_false(
                MAX_RAM >= int(values["memory"]) > MIN,
                f"RAM must be {MIN} and {MAX_RAM}.",
            )
            values["memory"] = int(values["memory"])
            values["image_name"] = (
                values["image_name"]
                or values["code"] + ":" + values["version"]
            )
            return values
        except AssertionError as e:
            messagebox.showerror("Error", f"AN EXCEPTION OCCURRED! {e}")
            return False

    # window
    window = ttk.Window(themename="darkly")
    window.resizable(True, True)
    window.title("Tool Publishing GUI")
    window.geometry("600x800")

    write_dir = ttk.StringVar()

    # --- SCROLLBAR & CANVAS SETUP ---

    # Main container to hold canvas and scrollbar
    main_container = ttk.Frame(window)
    main_container.pack(fill=BOTH, expand=YES)
    # Canvas
    canvas = ttk.Canvas(main_container)
    # Scrollbar linked to canvas
    vbar = ttk.Scrollbar(main_container, orient="vertical", command=canvas.yview)
    canvas.configure(yscrollcommand=vbar.set)

    vbar.pack(side=RIGHT, fill=Y)
    canvas.pack(side=LEFT, fill=BOTH, expand=YES)

    # The actual frame that holds the content
    content_frame = ttk.Frame(canvas, padding=20)
    canvas_window_id = canvas.create_window((0, 0), window=content_frame, anchor="nw", tags="inner_frame")

    # --- RESIZE LOGIC ---

    # Update scrollregion when content changes size
    def configure_scroll_region(event):
        canvas.configure(scrollregion=canvas.bbox("all"))

    content_frame.bind("<Configure>", configure_scroll_region)

    # Force the inner frame to match the canvas width (Horizontal Fill)
    def on_canvas_resize(event):
        canvas.itemconfig(canvas_window_id, width=event.width)

    canvas.bind("<Configure>", on_canvas_resize)

    def _on_mousewheel_windows(event):
        # Windows/macOS: event.delta is usually multiples of 120
        # We divide by 120 to get the number of "steps"
        canvas.yview_scroll(int(-1 * (event.delta / 120)), "units")

    def _on_linux_scroll_up(event):
        # Linux (X11) often sends Button-4 for scroll up
        canvas.yview_scroll(-1, "units")

    def _on_linux_scroll_down(event):
        # Linux (X11) often sends Button-5 for scroll down
        canvas.yview_scroll(1, "units")

    # Bind Standard MouseWheel (Windows/macOS)
    window.bind_all("<MouseWheel>", _on_mousewheel_windows)
    # Bind Linux specific buttons (X11)
    window.bind_all("<Button-4>", _on_linux_scroll_up)
    window.bind_all("<Button-5>", _on_linux_scroll_down)

    # --- WIDGETS ---

    # Logo Image
    try:
        # Adjusted path logic for safety
        img_path = os.path.join(os.path.dirname(__file__), "templates_tool_maker", "qmenta.png")
        if os.path.exists(img_path):
            logo = Image.open(img_path).resize((500, 110))
            img = ImageTk.PhotoImage(logo)
            image_label = ttk.Label(master=content_frame, image=img, anchor=CENTER)
            image_label.image = img  # Keep reference
            image_label.pack(side=TOP, pady=(0, 20))
    except Exception as e:
        print(f"Image load failed: {e}")

    # Title
    ttk.Label(
        master=content_frame,
        text="QMENTA Platform credentials",
        font="Helvetica 16 bold",
    ).pack(pady=10)

    ttk.Label(
        master=content_frame,
        text="(*) indicates mandatory field.",
        bootstyle="warning"
    ).pack(pady=(0, 20))

    # Helper to create rows easily
    def create_row(parent, label_text, is_password=False, default=None):
        frame = ttk.Frame(master=parent)
        frame.pack(fill=X, pady=5)

        lbl = ttk.Label(master=frame, text=label_text, width=30)
        lbl.pack(side=LEFT)

        entry = ttk.Entry(master=frame, show="*" if is_password else None)
        if default:
            entry.insert(0, str(default))
        entry.pack(side=LEFT, fill=X, expand=YES, padx=10)
        return entry

    # User inputs
    user_id_entry = create_row(content_frame, "Username*")
    password_entry = create_row(content_frame, "Password*", is_password=True)

    # Browse Folder Row
    folder_frame = ttk.Frame(master=content_frame)
    folder_frame.pack(fill=X, pady=10)
    ttk.Label(folder_frame, text="Select tool folder*", width=30).pack(side=LEFT)
    ttk.Button(folder_frame, text="Browse Folder", command=browse_folder).pack(side=LEFT, padx=10)
    ttk.Label(folder_frame, textvariable=write_dir).pack(side=LEFT, padx=10)

    # Tool info
    tool_id_entry = create_row(content_frame, "Specify the tool ID.*")
    tool_version_entry = create_row(content_frame, "Specify the tool version.*")
    tool_name_entry = create_row(content_frame, "Specify the tool name.*")
    num_cores_entry = create_row(content_frame, f"Cores required (integer, max {MAX_CORES})", default="1")
    memory_entry = create_row(content_frame, f"RAM required GB (integer, max {MAX_RAM})", default="1")
    image_name_entry = create_row(content_frame, "Docker image name")

    # Separator
    ttk.Separator(content_frame, orient=HORIZONTAL).pack(fill=X, pady=20)

    # Docker registry
    ttk.Label(
        master=content_frame,
        text="Docker Registry Credentials",
        font="Helvetica 12 bold"
    ).pack(pady=10)
    docker_url_entry = create_row(content_frame, "Docker registry URL", default="hub.docker.com")
    docker_user_entry = create_row(content_frame, "Docker registry user")
    docker_password_entry = create_row(content_frame, "Docker registry password*", is_password=True)

    # --- ACTIONS ---

    gui_content = {}

    def submit_form():
        gui_content.update(generate_output_object())
        gui_content.update(perform_selection_check(gui_content))
        if gui_content:
            window.destroy()

    action_frame = ttk.Frame(master=content_frame)
    action_frame.pack(pady=30)
    ttk.Button(
        master=action_frame,
        text="Publish in QMENTA Platform",
        command=submit_form,
        bootstyle="primary",
    ).pack(side=LEFT, pady=10)

    def quit_tool_creator():
        window.destroy()
        print("GUI closed.")
        exit()

    ttk.Button(
        action_frame,
        text="Quit",
        command=quit_tool_creator,
        bootstyle="danger",
    ).pack(side=LEFT, padx=10)

    # run
    window.mainloop()
    return gui_content


def main():
    content_build = gui_tkinter()
    if not content_build:
        exit()
    os.chdir(dir_name)

    user = content_build["qmenta_user"]
    password = content_build["qmenta_password"]
    try:
        auth = platform.Auth.login(
            username=user,
            password=password,
            base_url="https://platform.qmenta.com",
            ask_for_2fa_input=False,
        )
    except Needs2FAError as needs2faerror:
        messagebox.showinfo(
            "Message",
            str(needs2faerror)
            + " Please check the terminal to add the code sent to your phone.",
        )
        auth = platform.Auth.login(
            username=user,
            password=password,
            base_url="https://platform.qmenta.com",
            code_2fa=input("Input 2-FA code:"),
        )

    # Get information from the advanced options file.
    advanced_options = "settings.json"
    raise_if_false(
        os.path.exists(advanced_options),
        "Settings do not exist! Run the local test to create it.",
    )
    with open(advanced_options) as fr:
        content_build["advanced_options"] = fr.read()

    # Get information from the description file.
    with open("description.html") as fr:
        content_build["description"] = fr.read()

    with open("results_configuration.json", "r") as file:
        results_config = json.load(file)
    # The screen value is expected as a string with escaped chars (dict)
    results_config["screen"] = json.dumps(results_config["screen"])
    content_build["results_configuration"] = json.dumps(
        results_config
    ).replace("{}", "")

    content_build.update(
        {
            "start_condition_code": "output={'OK': True, 'code': 1}",
            "entry_point": "/root/entrypoint.sh",
            "tool_path": "tool:run",
        }
    )

    # After creating the workflow, the ID of the workflow must be requested
    # and added to the previous dictionary
    # otherwise it will keep creating new workflows on the platform
    # creating conflicts.
    res = platform.post(
        auth, "analysis_manager/upsert_user_tool", data=content_build
    )

    if res.json()["success"] == 1:
        print("Tool updated successfully!")
        print(
            "Tool name:",
            content_build["name"],
            "(",
            content_build["code"],
            ":",
            content_build["version"],
            ")",
        )
    else:
        print("ERROR setting the tool.")
        print(res.json())


if __name__ == "__main__":
    main()



import glob
import inspect
import json
import logging
import os
import shlex
import shutil
import subprocess
from collections.abc import Mapping
from types import SimpleNamespace
from typing import List, Union

from qmenta.sdk.context import AnalysisContext
from qmenta.sdk.local.context import LocalAnalysisContext
from qmenta.sdk.local.parse_settings import parse_tool_settings
from .context import TestFileInput
from .file_filter import ConditionForFilter
from .inputs import (
    CheckBox,
    ContainerHandler,
    Decimal,
    Heading,
    IndentText,
    InfoText,
    Integer,
    Line,
    MultipleChoice,
    Options_type,
    SingleChoice,
    String,
    SubjectHandler,
)
from .make_files import build_local_dockerfile, build_local_requirements
from .modalities import Modality, Tag
from .run_test_docker import run_docker


def add_container_filters(file_list: List) -> str:
    final_file_filter = list()
    or_file_filter = list()
    for idx, inp in enumerate(file_list):
        if not isinstance(inp, list):
            conditions = list()
            if not isinstance(inp.filter_file, list):
                inp.filter_file = [inp.filter_file]

            for idf, inf in enumerate(inp.filter_file):
                filter_file_conditions = list()
                if inf.modality != Modality.none:
                    filter_file_conditions.append(inf.modality)
                if inf.tags:
                    filter_file_conditions.append(inf.tags)
                if inf.regex != ".*":
                    filter_file_conditions.append(inf.regex)
                conditions.append((filter_file_conditions, idf))

            file_filter = ConditionForFilter(
                inp.id,
                conditions,
                min_files=inp.min_files,
                max_files=inp.max_files,
            )
            final_file_filter.append(
                f"{inp.id}{file_filter}"
                if inp.mandatory
                else f"({inp.id}{file_filter} OR TRUE)"
            )

        else:
            and_section = add_container_filters(inp)
            if idx > 0 and isinstance(file_list[idx - 1], list):
                or_file_filter.append(and_section)
            else:
                or_file_filter = list()
                final_file_filter.append(and_section)

    if or_file_filter:
        [final_file_filter.append(el) for el in or_file_filter]
        final_file_filter = "(" + " OR ".join(final_file_filter) + ")"
        return final_file_filter
    else:
        final_file_filter = (
            " AND ".join(final_file_filter)
            if len(file_list) == 1
            else "(" + " AND ".join(final_file_filter) + ")"
        )
        return final_file_filter


def create_setting_values_json(
    args: SimpleNamespace, input_folder: str, settings=None
):
    if settings is None:
        settings = []

    class MyEncoder(json.JSONEncoder):
        def default(self, o):
            if isinstance(o, Mapping):
                return {}
            if not isinstance(o, TestFileInput):
                print(o)
                return o.__dict__
            attrs = [
                "path",
                "modality",
                "tags",
                "file_filter_condition_name",
                "file_info",
            ]
            return {a: getattr(o, a) for a in attrs if bool(getattr(o, a))}

    outfile = os.path.join(input_folder, "setting_values.json")
    encoded = json.dumps(args.__dict__, indent=2, cls=MyEncoder)
    if settings:
        for k, v in args.__dict__.items():
            if (
                k != "test_name"
                and k != "sample_data_folder"
                and not sum([s["id"] == k for s in settings if "id" in s]) == 1
            ):
                raise KeyError(
                    f"Key {k} defined in test arguments, not found in settings."
                )
    with open(outfile, "w") as f:
        f.write(encoded)
    return outfile


class FilterFile:
    """
    By specifying file filters, tools can provide usage information to the user regarding the expected
    type of file(s) that it needs to start the processing. This can include, for instance, the modality of the
    image or some required tags. File filters are also useful when we want the user to select different types of
    files within one input container.

    Parameters
    ----------
    modality: Modality or list
        Modality of the sequence required by the tool. Check the qmenta.sdk.tool_maker.modalities.Modality list
        to check the possible values.
    tags: list
        Tags of the sequence required by the tool.
        Example: Tag("mask"), Tag("labels"). It is equivalent to string.
    regex: str
        Regular expression to identify by file name.
    """

    def __init__(
        self,
        modality: Union[Modality, List[Modality]] = Modality.none,
        tags: List[Tag] = None,
        regex: str = ".*",
    ):
        if tags is None:
            tags = list()
        self.modality = modality
        self.tags = tags
        self.regex = regex


class InputFile:
    """
    Defines one or more input files required by the tool to run. Here is specified which conditions the file/s
    should adhere to in order to be selected to be processed. The filter defines the modality, tag and/or file name
    regular expression the desired file should have.

    Parameters
    ----------
    file_filter_condition_name: str
        Is the ID that will be used to define the file filter condition and also used
        to get the file in the tool from the self.inputs.{container_id}.{file_filter_condition_name} dictionary
    filter_file
        Definition of the input requirement.
    mandatory
        Declares that the file must exist in the inputs when is set to 1. Optional when is set to 0.
    min_files
        Specify the minimum number of files that are accepted using this filter.
    max_files
        Specify the maximum number of files that are accepted using this filter.
    """

    def __init__(
        self,
        file_filter_condition_name: str,
        filter_file: Union[FilterFile, List[FilterFile]] = FilterFile(
            regex=".*"
        ),
        mandatory: int = 1,
        min_files: int = 1,
        max_files: Union[int, str] = "*",
    ):
        self.id = file_filter_condition_name
        self.filter_file = filter_file
        self.mandatory = mandatory
        self.min_files = min_files
        self.max_files = max_files


class Tool:
    """
    Class which defines the tool and its components.

    Parameters
    ----------
    context: AnalysisContext
        A QMENTA context instance
    testing_tool: bool
        Flag activated when the tool is being run locally for testing.
    """

    def __init__(self, context: AnalysisContext = None, testing_tool=False):

        self.context = context
        self.inputs = SimpleNamespace()
        self.testing = True if context is None else False
        self._inputs = []
        self.testing_tool = testing_tool

        self.tool_inputs()
        self.tool_path = os.path.dirname(inspect.getfile(self.__class__))
        self.settings_path = os.path.join(self.tool_path, "settings.json")

    def tool_inputs(self):
        """
        Mandatory method in the Tools class. It must contain the input requirements.
        """
        if not self.testing_tool:
            raise NotImplementedError(
                "You forgot to define the inputs in tool_inputs"
            )

    def run(self, context=None):
        """
        Mandatory method in the Tools class. It must contain the processing script.
        """
        raise NotImplementedError("You forgot to implement the run method")

    def tool_outputs(self):
        """
        Optional method in the Tools class. It will contain the specifications of the tool's result configuration.
        """
        pass

    def add_input_container(
        self,
        container_id: str,
        file_list: Union[List[InputFile], List[List[InputFile]]],
        title: str = "",
        info: str = "",
        anchor: int = 1,
        batch: int = 1,
        mandatory: int = 1,
        tool_codes_set_as_input=None,
    ):
        """
        By default, the settings configuration of a tool includes a generic input container specification
        that allows any kind of data to be selected as the input for a tool. In this context, an input container
        is just a group of files in a specific structure. The most typical examples of input containers are a patient’s
        data for a single session or the result of an analysis.

        For more information please visit:
        https://docs-dev.qmenta.com/sdk/guides_docs/settings.html#handling-the-input-files

        Parameters
        ----------
        container_id: str
            A unique identifier for a session or input container.
        file_list: list
            a list of objects `InputFile`
        title: str
            Title displayed in the platform for the input.
        info: str
            Information text displayed in the platform for the input.
        anchor: int
        batch: int
            allows the tool can to be launched in parallel for multiple sessions
            or analysis results. That is, perform a batch execution.
        mandatory: int, 0 or 1
            indicates whether the user must provide data for this input container.
        tool_codes_set_as_input: list or None
            List[String] a list of tools identifiers to use their results as input data (e.g. qmenta_acpc_alignment).

        """
        if tool_codes_set_as_input is None:
            tool_codes_set_as_input = list()
        _final_file_filter = add_container_filters(file_list)
        container = ContainerHandler(
            id_=container_id,
            file_filter=_final_file_filter,
            title=title,
            info=info,
            anchor=anchor,
            batch=batch,
            mandatory=mandatory,
            tool_codes_set_as_input=tool_codes_set_as_input,
        )
        self._inputs.append(container)

    def add_subject(
        self,
        container_id: str,
        file_list: Union[List[InputFile], List[List[InputFile]]],
        title: str = "",
        info: str = "",
        anchor: int = 1,
        batch: int = 1,
        mandatory: int = 1,
        tool_codes_set_as_input=None,
    ):
        if tool_codes_set_as_input is None:
            tool_codes_set_as_input = list()
        _final_file_filter = add_container_filters(file_list)
        self._inputs.append(
            SubjectHandler(
                id_=container_id,
                file_filter=_final_file_filter,
                title=title,
                info=info,
                anchor=anchor,
                batch=batch,
                mandatory=mandatory,
                tool_codes_set_as_input=tool_codes_set_as_input,
            )
        )

    def add_input_single_choice(
        self, id_: str, options: Options_type, default: object, title: str = ""
    ):
        """
        Dropdown menu
        https://docs.qmenta.com/sdk/guides_docs/settings.html#single-choice-parameter

        Parameters
        ----------
        id_: str
            Identifier of the parameter, should be unique in the settings for the tool
        options: list
            The format must be the following: [("a", "Option A"), ("b", "Option B"), ("c", "Option C")]
            Each element is a two-element tuple where the first element is the value id and the second is the
            value representation. The first element is the value returned to the tool and the second is the one
            displayed in the platform
        default: str
            Must be a value id from the options defined. Is going to be the one selected by default
            when opening the tool in the platform.
        title: str
            text shown besides the parameter to shortly describe the parameter function

        """
        opt = SingleChoice(id_, options, default, title)
        if default not in [o[0] for o in options]:
            raise ValueError(
                f"Default value [{default}] not in options {[o[0] for o in options]}"
            )
        self._inputs.append(opt)
        setattr(self.inputs, id_, opt)

    def add_input_multiple_choice(
        self,
        id_: str,
        options: Options_type,
        default: List[str],
        title: str = "",
    ):
        """
        https://docs.qmenta.com/sdk/guides_docs/settings.html#multiple-choice-parameter

        Parameters
        ----------
        id_: str
            Identifier of the parameter, should be unique in the settings for the tool
        options: list
            The format must be the following: [("a", "Option A"), ("b", "Option B"), ("c", "Option C")]
            Each element is a two-element tuple where the first element is the value id and the second is the
            value representation. The first element is the value returned to the tool and the second is the one
            displayed in the platform
        default: list
            Must be a list of the value ids defined in the options. Is going to be the ones selected by default
            when opening the tool in the platform.
        title: str
            text shown besides the parameter to shortly describe the parameter function

        """
        opt = MultipleChoice(id_, options, default, title)
        for defa in default:
            if defa not in [o[0] for o in options]:
                raise ValueError(
                    f"Default value ({defa}) not in options {[o[0] for o in options]}"
                )
        self._inputs.append(opt)
        setattr(self.inputs, id_, opt)

    def add_input_checkbox(self, id_: str, default: int, title: str = ""):
        """
        https://docs.qmenta.com/sdk/guides_docs/settings.html#checkbox-parameter

        Parameters
        ----------
        id_: str
            Identifier of the parameter, should be unique in the settings for the tool
        default: 0 or 1
            Is going to be the value selected by default when opening the tool in the platform.
        title: str
            text shown besides the parameter to shortly describe the parameter function

        """
        self._inputs.append(CheckBox(id_, title, default))

    def add_input_string(self, id_: str, default: str, title: str = ""):
        """
        https://docs.qmenta.com/sdk/guides_docs/settings.html#string-parameter

        Parameters
        ----------
        id_: str
            Identifier of the parameter, should be unique in the settings for the tool
        default: str
            Is going to be the value selected by default when opening the tool in the platform.
        title: str
            text shown besides the parameter to shortly describe the parameter function

        """
        self._inputs.append(String(id_, title, default))

    def add_input_decimal(
        self,
        id_: str,
        default: float,
        title: str = "",
        minimum: float = -1e9,
        maximum: float = 1e9,
    ):
        """
        https://docs.qmenta.com/sdk/guides_docs/settings.html#decimal-parameter

        Parameters
        ----------
        id_: str
            Identifier of the parameter, should be unique in the settings for the tool
        default: float
            Is going to be the value selected by default when opening the tool in the platform.
        title: str
            text shown besides the parameter to shortly describe the parameter function
        minimum: float
            Smallest value that can be stored in this parameter
        maximum: float
            Highest value that can be stored in this parameter

        """
        self._inputs.append(Decimal(id_, title, default, minimum, maximum))

    def add_input_integer(
        self,
        id_: str,
        default: int,
        title: str = "",
        minimum=0,
        maximum=1e9,
    ):
        """
        https://docs.qmenta.com/sdk/guides_docs/settings.html#integer-parameter

        Parameters
        ----------

        id_: str
            Identifier of the parameter, should be unique in the settings for the tool
        default: int
            Is going to be the value selected by default when opening the tool in the platform.
        title: str
            text shown besides the parameter to shortly describe the parameter function
        minimum: float
            Smallest value that can be stored in this parameter
        maximum: float
            Highest value that can be stored in this parameter

        """
        self._inputs.append(Integer(id_, title, default, minimum, maximum))

    """
    The following methods are texts and separators.
    Learn more: https://docs-dev.qmenta.com/sdk/guides_docs/settings.html#adding-text-and-separators
    """

    def add_line(self):
        """
        Learn more:  https://docs-dev.qmenta.com/sdk/guides_docs/settings.html#line

        Adds a horizontal line to divide settings' sections.
        """
        self._inputs.append(Line())

    def add_heading(self, content: str):
        """
        Learn more: https://docs-dev.qmenta.com/sdk/guides_docs/settings.html#heading-text

        Parameters
        ----------
        content: str
            Message to be shown in the header as a text in the platform

        """
        self._inputs.append(Heading(content))

    def add_info(self, content: str):
        """
        Learn more:  https://docs-dev.qmenta.com/sdk/guides_docs/settings.html#info-text

        Parameters
        ----------
        content: str
            Message to be shown as an information box in the platform

        """
        self._inputs.append(InfoText(content))

    def add_indent(self, content: str):
        """
        Learn more:  https://docs-dev.qmenta.com/sdk/guides_docs/settings.html#indented-text

        Parameters
        ----------
        content: str
            Message to be shown in the as an indented text in the platform


        """
        self._inputs.append(IndentText(content))

    def prepare_inputs(
        self,
        context: Union[AnalysisContext, LocalAnalysisContext],
        logger: logging.Logger,
    ):
        """
        Prepare the inputs for the tool. Downloads the data and sets input variables for the tool.

        Parameters
        ----------
        context: AnalysisContext or LocalAnalysisContext
            The context of the analysis given by the SDK
        logger: logging.Logger
            The logger to use for logging

        """
        context.fetch_analysis_data()
        for inp in self._inputs:
            if (
                isinstance(inp, Line)
                or isinstance(inp, Heading)
                or isinstance(inp, InfoText)
                or isinstance(inp, IndentText)
            ):
                continue
            try:
                setattr(self.inputs, inp.id, inp.get_input(context))
            except Exception as e:
                logger.error(f"Could not download input: {e}")
                if inp.mandatory:
                    raise
                setattr(self.inputs, inp.id, None)

    def generate_settings_file(self) -> list:
        """
        Generates a settings file for the tool.
        """
        settings = [iw.__dict__ for iw in self._inputs]
        encoded = json.dumps(settings, indent=2)
        if not self.testing_tool:
            with open(self.settings_path, "w") as f:
                f.write(encoded)
        return settings

    def copy_input_files_to_folder(self, args: dict) -> SimpleNamespace:
        """
        Used in local testing.
        Copies all input files to a folder.

        Parameters
        ----------
        args: SimpleNamespace
            The arguments passed to the tool.

        """
        logger = logging.getLogger(__name__)
        input_folder = os.path.join(
            self.tool_path, "local", "test", args["test_name"], "input_folder"
        )
        sample_data_folder = os.path.join(
            self.tool_path, "local", "test", args["sample_data_folder"]
        )
        tools_resources = glob.glob(
            os.path.join(
                self.tool_path, "local", "test", "tools_resources", "**", "*.*"
            ),
            recursive=True,
        )
        execution_folder = os.path.join(
            self.tool_path,
            "local",
            "test",
            args["test_name"],
            "execution_folder",
        )
        for res in tools_resources:
            rel_path = os.path.relpath(
                res, os.path.join(self.tool_path, "local", "test")
            )
            os.makedirs(
                os.path.dirname(os.path.join(execution_folder, rel_path)),
                exist_ok=True,
            )
            shutil.copy(res, os.path.join(execution_folder, rel_path))

        new_args = args
        for id_key in args:
            ref_test_file = None
            argument_files = None
            mandatory = None
            if isinstance(args[id_key], dict) and "files" in args[id_key]:
                argument_files = args[id_key]["files"]
                mandatory = args[id_key]["mandatory"]
                ref_test_file = args[id_key]["files"][0]
            elif isinstance(args[id_key], list) and isinstance(
                args[id_key][0], TestFileInput
            ):
                # TO DEPRECATE THIS CASE in the near future
                argument_files = args[id_key]
                mandatory = (
                    True  # before this was not taken into account, always True
                )
                ref_test_file = args[id_key][0]
            if ref_test_file is None:
                continue

            new_files_id = list()
            failed_copy_files = 0
            is_one_mandatory = False
            for lfile in argument_files:
                is_one_mandatory = is_one_mandatory or lfile.mandatory
                if os.path.isdir(os.path.join(sample_data_folder, lfile.path)):
                    dst = os.path.join(input_folder, id_key, lfile.path)
                    os.makedirs(os.path.dirname(dst), exist_ok=True)
                    src = os.path.join(sample_data_folder, lfile.path)
                    for filename in glob.glob(
                        os.path.join(src, "**"), recursive=True
                    ):
                        if os.path.isdir(filename):
                            # skip folder
                            continue
                        rel_path = os.path.relpath(
                            filename,
                            os.path.join(sample_data_folder, lfile.path),
                        )
                        dst = os.path.join(
                            input_folder, id_key, lfile.path, rel_path
                        )
                        src = filename
                        os.makedirs(os.path.dirname(dst), exist_ok=True)
                        # Copy to input folder
                        try:
                            shutil.copy(src, dst)
                        except FileNotFoundError:
                            if mandatory:
                                raise
                            logger.error(
                                f"File not found in optional argument: {src}"
                            )
                            failed_copy_files += 1
                            break

                        # Include file in args. Modality is not considered
                        newfileinput = TestFileInput(
                            path=os.path.join(lfile.path, rel_path),
                            file_name_path=os.path.join(lfile.path, rel_path),
                            file_filter_condition_name=ref_test_file.file_filter_condition_name,
                            file_info=ref_test_file.file_info,
                            tags=ref_test_file.tags,
                            regex=ref_test_file.regex,
                            mandatory=ref_test_file.mandatory,
                        )
                        new_files_id.append(newfileinput)

                else:
                    # Copy the source files
                    dst = os.path.join(input_folder, id_key, lfile.path)
                    os.makedirs(os.path.dirname(dst), exist_ok=True)
                    src = os.path.join(sample_data_folder, lfile.path)
                    try:
                        shutil.copy(src, dst)
                    except FileNotFoundError:
                        failed_copy_files += 1
                        if (
                            mandatory
                            and len(argument_files) == failed_copy_files
                        ):
                            raise ValueError(
                                f"All files not found in mandatory argument {id_key}"
                            )

                        logger.error(
                            f"File not found in optional argument: {src}"
                        )
                    new_files_id.append(lfile)

            # if files in the defined folder in test where added as TestFileInput, replace the value in the
            # setting id so the download gets the files instead of the folder.
            # setting id so the download gets the files instead of the folder.
            new_args[id_key] = new_files_id
            if mandatory and not is_one_mandatory:
                raise Exception(
                    f"The setting input {id_key} is set to mandatory but none of the condition filters is mandatory. "
                    f"At least one should be mandatory. Make sure you have defined the 'mandatory' argument in the"
                    f" 'TestFileInput' of you local test"
                )
            elif not mandatory and is_one_mandatory:
                raise Exception(
                    f"The setting input {id_key} is NOT set to mandatory and there is one condition filter is set to "
                    f"mandatory. All filters in this input should be set NOT mandatory. Make sure you have defined "
                    f"the 'mandatory' argument in the 'TestFileInput' of you local test"
                )

        return SimpleNamespace(**new_args)

    def test_with_args(
        self,
        in_args: dict,
        overwrite_settings: bool = True,
        refresh_test_data: bool = False,
    ) -> None:
        """
        Run the tool with the given arguments.

        Parameters
        ----------
        in_args: dict
            The arguments to pass to the tool.
        overwrite_settings: bool
            If True, the settings file will be overwritten with the current settings.
        refresh_test_data: bool
            If True, the test data will be removed, and it will create the folder structure and copy the files again.
        """
        args = SimpleNamespace(**in_args)
        local_test_folder = os.path.join(
            self.tool_path, "local", "test", args.test_name
        )

        settings = dict()
        if overwrite_settings:
            settings = self.generate_settings_file()
        if refresh_test_data and os.path.exists(local_test_folder):
            shutil.rmtree(local_test_folder)

        current_dir = os.getcwd()  # Remember the current directory
        # Create input folder
        input_folder = os.path.join(local_test_folder, "input_folder")
        os.makedirs(input_folder, exist_ok=True)

        # Create execution folder
        exec_folder = os.path.join(local_test_folder, "execution_folder")
        os.makedirs(exec_folder, exist_ok=True)

        # Create output folder
        out_folder = os.path.join(local_test_folder, "output_folder")
        os.makedirs(out_folder, exist_ok=True)

        # Copy input files to input folder and update settings_values for
        # folders
        input_settings = os.path.join(input_folder, "setting_values.json")
        if refresh_test_data or not os.path.exists(input_settings):
            args = self.copy_input_files_to_folder(in_args)

            # Create settings_values.json in input_folder
            input_settings = create_setting_values_json(
                args, input_folder, settings
            )

        # Parse settings
        settings = parse_tool_settings(self.settings_path, input_settings)

        # Instantiate local context
        context = LocalAnalysisContext(settings, input_folder, out_folder, "")

        # Change to execution folder
        os.environ["MINTEXE_PATH"] = exec_folder

        # Run the tool
        logger = logging.getLogger("main")
        logger.setLevel(logging.INFO)
        file_ch = logging.FileHandler("logger.log", "w")
        file_ch.setLevel(logging.INFO)
        logger.addHandler(file_ch)
        logger.info("This is the execution log --> Starting context.")

        self.tool_outputs()
        self.run(context)

        # Back to the original directory
        os.chdir(current_dir)

    def test_docker_with_args(
        self,
        version: str,
        in_args: dict,
        stop_container: bool = True,
        delete_container: bool = True,
        attach_container: bool = True,
    ):
        # Parse in_args as SimpleNameSpace
        # Load the requirements from the build yaml file
        local_folder = os.path.join(self.tool_path, "local")
        os.chdir(
            local_folder
        )  # move here to create the Dockerfile and requirements.txt

        build_local_requirements()
        build_local_dockerfile()

        os.chdir(self.tool_path)
        args = SimpleNamespace(**in_args)
        # Build the Docker image found in the Dockerfile
        local_tests_folder = os.path.join(local_folder, "test")
        local_test_folder = os.path.join(local_tests_folder, args.test_name)
        # run the docker build commandline
        base_image_name = os.path.basename(self.tool_path) + ":" + version
        # copy the tool.py file in the tool path to the local folder
        shutil.copy(
            os.path.join(self.tool_path, "tool.py"),
            os.path.join(self.tool_path, "local"),
        )
        docker_build_command = f"docker build -t {base_image_name} ."
        print(
            f"Building the docker image {os.path.basename(self.tool_path)}..."
        )
        print(docker_build_command)

        os.chdir(local_folder)
        process = subprocess.Popen(
            shlex.split(docker_build_command),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        _, error_data = process.communicate(
            timeout=99999
        )  # Halts execution until process finishes
        if process.returncode:
            raise OSError(error_data)
        # The following folders should already exist because we have run the tool locally, otherwise create them
        # Create input folder
        input_folder = os.path.join(local_test_folder, "input_folder")
        os.makedirs(input_folder, exist_ok=True)

        # Create execution folder
        exec_folder = os.path.join(local_test_folder, "execution_folder")
        os.makedirs(exec_folder, exist_ok=True)

        # Create output folder
        out_folder = os.path.join(local_test_folder, "output_folder")
        os.makedirs(out_folder, exist_ok=True)
        run_docker(
            image=base_image_name,
            inputs=input_folder,
            outputs=out_folder,
            overwrite_output=True,
            settings=os.path.join(self.tool_path, "settings.json"),
            values=os.path.join(input_folder, "setting_values.json"),
            mounts=[f"{self.tool_path}/tool.py:/qmenta/tool.py"],
            resources=os.path.join(local_tests_folder),
            stop_container=stop_container,
            delete_container=delete_container,
            attach_container=attach_container,
        )



#!/usr/bin/env python

"""
Basic QMENTA Tool Image Local Tester

Example:

$ python test_sdk_tool.py -v /home/apuente/repos/qmenta_imaging/tools/lesion_toads/tool.py:/qmenta/tool.py \
    gcr.io/hosting-setup-853/lesion_toads_neuro-269:1.0.366 \
    ~/data/lesion_toads/ \
    ~/data/output/ \
    --settings ./tools/lesion_toads/settings.json \
    --values ./tools/lesion_toads/mock_values.json

Usage:
 $ ./test_sdk_tool.py image_name inputs/ outputs
 $ ./test_sdk_tool.py image_name inputs/ outputs --settings settings.json --values values.json --tool package.tool -v \
    /host/path:/container/path

Compatible with Python 3 with no additional packages.
See https://docs.qmenta.com/sdk/testing.html for more information.
"""

import os
import random
import string
import subprocess
from os import listdir
from os.path import abspath, isdir

import sys


def random_seq(n):
    """
    Generates a random sequence of upper-cased characters of size n
    """
    return "".join(random.choice(string.ascii_uppercase) for x in range(n))


def run_command(cmd, verbose=True):
    """
    Execute command in shell
    """
    if verbose:
        print(" ".join(cmd))
    subprocess.call(cmd)


def error(msg):
    """
    Exit with error
    """
    print(msg)
    sys.exit(1)


def run_docker(
    image: str,
    inputs: str,
    outputs: str,
    settings: str,
    values: str,
    mounts: list,
    overwrite_output: bool,
    resources: str,
    stop_container: bool,
    delete_container: bool,
    attach_container: bool,
):
    """
    Script entrypoint
    """

    if not isdir(abspath(inputs)):
        error("Error: Input folder does not exist")
    if not isdir(abspath(outputs)):
        error("Error: Output folder does not exist")
    if not os.listdir(abspath(outputs)) == []:
        print(
            "Warning: Output folder is not empty (files could be overwritten)."
        )
        if not overwrite_output:
            sys.exit(1)
    if resources and not isdir(abspath(resources)):
        error("Error: Resource file does not exist")
    if settings and not values:
        error(
            "Error: Entering a custom settings file requires a settings values file"
        )

    # In-container paths
    c_settings_path = os.path.join("/", "qmenta", "local_exec_settings.json")
    c_values_path = os.path.join(
        "/", "qmenta", "local_exec_settings_values.json"
    )
    c_input_path = os.path.join("/", "qmenta", "local_exec_input/")
    c_output_path = os.path.join("/", "qmenta", "local_exec_output/")
    c_res_path = os.path.join("/", "qmenta", "local_exec_resources/")

    # If no tool_settings file is provided, use a generic one
    if not settings:
        settings_content = """[
    {
        "type": "container",
        "title": "Example Container",
        "id":"input",
        "mandatory":1,
        "batch":1,
        "file_filter": "c_files[1,*]<'', [], '.*'>",
        "in_filter":["mri_brain_data"],
        "out_filter":[],
        "anchor":1
    }
]"""

        settings = "./generic_settings.json"

        with open(settings, "w") as settings_file:
            settings_file.write(settings_content)

    # If no tool_settings_values file is provided, generate just the list of input files
    if not values:
        files = listdir(os.path.join(abspath(inputs), "input"))
        values_content = (
            '{"input":[\n'
            + ",\n".join(
                ['   {"path": "' + f + '"}' for f in files if not isdir(f)]
            )
            + "]\n}"
        )

        values = "./generic_values.json"

        with open(values, "w") as values_file:
            values_file.write(values_content)

    # Extra directories to be mounted (for instance, live version of source code)
    extra_volumes = []
    for volume in mounts or []:
        extra_volumes += ["-v", volume]

    # Launch the detached container
    c_name = "local_test_" + random_seq(5)
    print("\nStarting container {}...".format(c_name))
    run_command(
        [
            "docker",
            "run",
            "-dit",
            "-v",
            abspath(inputs) + ":" + c_input_path,
            "-v",
            abspath(outputs) + ":" + c_output_path,
        ]
        + extra_volumes
        + [
            "--entrypoint=/bin/bash",
            "--name=" + c_name,
            image,
        ]
    )
    # Copy the settings files to the container
    run_command(
        ["docker", "cp", abspath(settings), c_name + ":" + c_settings_path],
        verbose=False,
    )
    run_command(
        ["docker", "cp", abspath(values), c_name + ":" + c_values_path],
        verbose=False,
    )

    # Changing to local executor
    run_command(
        [
            "docker",
            "exec",
            c_name,
            "/bin/bash",
            "-c",
            r"sed -i.bak 's/\<qmenta.sdk\>/&.local/' /qmenta/entrypoint.sh",
        ],
        verbose=False,
    )

    # Run the local executor
    print("\nRunning tool.py:run()...\n")
    launch_cmd = [
        "docker",
        "exec",
        c_name,
        "/bin/bash",
        "/qmenta/entrypoint.sh",
        c_settings_path,
        c_values_path,
        c_input_path,
        c_output_path,
        "--tool-path",
        "tool:run",
    ]

    print(launch_cmd)
    # Additional resources (QMENTA)
    if resources:
        run_command(
            ["docker", "cp", abspath(resources), c_name + ":" + c_res_path],
            verbose=False,
        )
        launch_cmd += ["--res-folder", c_res_path]
    run_command(launch_cmd)
    if stop_container:
        run_command(["docker", "stop", c_name])
    if delete_container:
        run_command(["docker", "rm", c_name], verbose=True)
    if attach_container:
        run_command(["docker", "attach", c_name], verbose=True)



#! /usr/bin/env python
import json
import os
import re
from tkinter import filedialog, CENTER, messagebox, END

import ttkbootstrap as ttk
from PIL import Image, ImageTk
from qmenta.core import platform
from qmenta.core.auth import Needs2FAError
from qmenta.sdk.tool_maker.make_files import raise_if_false

FONT = "Consolas"
FONTSIZE = 10

dir_name = ""


def gui_tkinter():
    """
    Creates a graphical user interface (GUI) for publishing workflows to the QMENTA platform.

    This function sets up a Tkinter-based GUI window with the following components:
    1. **Logo and Title**: Displays the QMENTA logo and a title indicating the purpose of the GUI.
    2. **User Credentials Input**: Fields for entering the QMENTA platform username and password.
    3. **Tool Information Input**: Fields for specifying the tool ID, version, and name.
    4. **Folder Selection**: A button to browse and select the folder where the tool is stored.
    5. **Action Buttons**: Buttons to submit the form (pushing the tool to the platform) or quit the application.

    The GUI collects user input, validates it, and returns a dictionary containing the collected data
    when the form is submitted.

    Returns
    -------
    dict
        A dictionary containing the following keys and values:
        - "qmenta_user" (str): The QMENTA platform username to log in.
        - "qmenta_password" (str): The QMENTA platform password to log in.
        - "code" (str): The tool ID.
            Your user ID will be prepended to the code to prevent overwriting existing workflows from other users.
        - "version" (str): The tool version.
        - "folder" (str): The directory path where the tool is stored.
        - "name" (str): The tool name or title.
            It will be the displayed name in the QMENTA Platform when you want to run it.
        - "short_name" (str): A processed version of the tool ID (lowercase, spaces replaced with underscores).

    Notes
    -----
    - The GUI uses the `ttkbootstrap` library for a modern and themed appearance.
    - The `browse_folder` function is called when the "Browse Folder" button is clicked. It updates the
      selected folder path and tool ID based on the folder name.
    - The `submit_form` function is called when the "Push to Platform" button is clicked. It validates
      the input data and closes the GUI if all validations pass.
    - The `quit_tool_creator` function is called when the "Quit" button is clicked. It closes the GUI
      and exits the application.

    Example
    -------
    >>> gui_data = gui_tkinter()
    # User interacts with the GUI:
    # - Enters username: "user123"
    # - Enters password: "password123"
    # - Selects folder: "/path/to/tool_folder"
    # - Enters tool ID: "my_tool"
    # - Enters tool version: "1.0.0"
    # - Enters tool name: "My Tool"
    # Clicks "Push to Platform".
    >>> print(gui_data)
    {
        "qmenta_user": "user123",
        "qmenta_password": "password123",
        "code": "my_tool",
        "version": "1.0.0",
        "folder": "/path/to/tool_folder",
        "name": "My Tool",
        "short_name": "my_tool"
    }
    """

    def browse_folder():
        """
        Handles the action triggered by clicking the "Browse" button to select a workflow folder.

        This method opens a directory dialog for the user to select a folder. Once a folder is selected,
        it performs the following actions:
        1. Sets the selected folder path in the `write_dir` variable.
        2. Updates the `tool_id_entry` field in the GUI with the name of the selected folder.
        3. Checks for a "version" file inside the selected folder. If the file exists, it reads the version
           number and updates the `tool_version_entry` field in the GUI.

        Notes
        -----
        - The `dir_name` variable is updated globally to store the selected folder path.
        - The `write_dir` variable is a `StringVar` GUI component that stores the directory path.
        - The `tool_id_entry` and `tool_version_entry` fields are GUI input fields that display the folder name
          and version, respectively.
        - If the "version" file does not exist in the selected folder, the `tool_version_entry` field remains
          unchanged.

        Example
        -------
        >>> browse_folder()
        # User selects the folder "/path/to/workflow_folder" which contains a "version" file with "1.0.0".
        # After execution:
        # - `dir_name` is set to "/path/to/workflow_folder".
        # - `tool_id_entry` displays "workflow_folder".
        # - `tool_version_entry` displays "1.0.0".
        """
        global dir_name
        dir_name = filedialog.askdirectory()
        write_dir.set(dir_name)
        tool_id_entry.delete(0, END)
        tool_id_entry.insert(0, os.path.basename(dir_name))
        # checking version
        version_file = os.path.join(dir_name, "version")
        if os.path.exists(version_file):
            with open(os.path.join(dir_name, "version")) as fv:
                version = fv.read()
            tool_version_entry.delete(0, END)
            tool_version_entry.insert(0, version)

    def generate_output_object():
        """
        This method collects data entered by the user in various input fields (e.g., user ID, password, tool ID, etc.)
        and organizes them into a structured dictionary. The dictionary can then be used for further processing,
        such as authentication, tool configuration, or saving settings.

        Returns
        -------
        dict
            A dictionary containing the following key-value pairs:
            - "qmenta_user" (str): The user ID to log in in the QMENTA Platform set in the GUI.
            - "qmenta_password" (str): The password to log in in the QMENTA Platform set in the GUI.
            - "code" (str): The tool ID set in the GUI.
            - "version" (str): The version of the tool set in the GUI.
            - "folder" (str): The directory path where outputs or files will be saved, as specified in the GUI.
            - "name" (str): The name of the tool set in the GUI.

        Example
        -------
        >>> result = generate_output_object()
        >>> print(result)
        {
            "qmenta_user": "user123",
            "qmenta_password": "password123",
            "code": "tool_abc",
            "version": "1.0",
            "folder": "/path/to/output",
            "name": "My Tool"
        }

        """
        values = {
            "qmenta_user": user_id_entry.get(),
            "qmenta_password": password_entry.get(),
            "code": tool_id_entry.get(),
            "short_name": tool_id_entry.get(),
            "version": tool_version_entry.get(),
            "folder": write_dir.get(),
            "name": tool_name_entry.get(),
        }
        return values

    def perform_selection_check(values):
        """
        Validates and processes the user input values collected from the GUI.

        This method performs a series of checks on the `values` dictionary to ensure that the input data
        is valid and meets the required criteria. If any validation fails, an error message
        is displayed, and the method returns `False`.

        Parameters
        ----------
        values : dict
            A dictionary containing user input values, typically generated by the `get_user_input_values`
            method. Expected keys include:
            - "code" (str): The tool ID.
            - "folder" (str): The directory path where outputs or files will be saved.
            - "name" (str): The name of the tool.
            - "version" (str): The version of the tool.

        Returns
        -------
        dict or bool
            - If all validations pass, the method returns the processed `values` dictionary with
              additional keys (e.g., "short_name") and modifications (e.g., lowercase tool ID).
            - If any validation fails, an error message is displayed using `messagebox.showerror`,
              and the method returns `False`.

        Raises
        ------
        AssertionError
            If any of the validation checks fail, an `AssertionError` is raised internally, and an
            error message is displayed to the user.

        Notes
        -----
        - The `values` dictionary is modified in place during processing:
            - The "code" (tool ID) must be lowercase and must not contain spaces.
            - The "folder" is validated to ensure it exists and is not empty.
            - The "short_name" key is added, derived from the "code" (tool ID) in lowercase and with
              spaces replaced by underscores.
        - The tool version is validated against a regex pattern to ensure it follows a valid format
          (e.g., "1.0", "2.1.0", "3.0.*").

        Example
        -------
        >>> input_user = {
        ...     "code": "My Tool",
        ...     "folder": "/path/to/folder",
        ...     "name": "My Tool Name",
        ...     "version": "1.p0"
        ... }
        >>> result = perform_selection_check(input_user)
        >>> print(result)
        False

        >>> input_user = {
        ...     "code": "my_tool",
        ...     "folder": "/path/to/folder",
        ...     "name": "My Tool Name",
        ...     "version": "1.0"
        ... }
        >>> result = perform_selection_check(input_user)
        >>> print(result)
        {
            "code": "my_tool",
            "short_name": "my_tool",
            "folder": "/path/to/folder",
            "name": "My Tool Name",
            "version": "1.0"
        }
        """
        try:
            # Validate and process the tool ID
            raise_if_false(
                isinstance(values["code"], str), "Tool ID must be a string."
            )
            raise_if_false(values["code"] != "", "Tool ID must be defined.")
            values["code"] = values["code"].lower()  # Convert to lowercase

            # Validate and process the folder
            raise_if_false(values["folder"] != "", "Folder must be defined.")
            raise_if_false(
                os.path.exists(values["folder"]), "Folder must exist."
            )

            # Ensure tool ID has no spaces and create a short name
            if " " in values["code"]:
                print(
                    "WARNING: Tool code can't have spaces, replacing with under score."
                )
                values["short_name"] = (
                    values["code"].lower().replace(" ", "_")
                )  # Create short name

            # Validate the tool name
            raise_if_false(
                isinstance(values["name"], str), "Tool name must be a string."
            )
            raise_if_false(values["name"] != "", "Tool name must be defined.")

            # Validate the tool version
            raise_if_false(
                values["version"] != "", "Tool version must be defined."
            )
            raise_if_false(
                re.search(r"^(\d+\.)?(\d+\.)?(\*|\d+)$", values["version"]),
                "Version format not valid.",
            )

            return values

        except AssertionError as e:
            messagebox.showerror("Error", f"AN EXCEPTION OCCURRED! {e}")
            return False

    # window
    window = ttk.Window(themename="darkly")
    window.title("Workflow Publishing GUI")

    write_dir = ttk.StringVar()
    # image
    logo = Image.open(
        os.path.join(
            os.path.dirname(__file__), "templates_tool_maker", "qmenta.png"
        )
    ).resize((500, 110))
    img = ImageTk.PhotoImage(logo)
    image_label = ttk.Label(master=window, image=img, padding=5, anchor=CENTER)
    image_label.config(image=img)
    image_label.pack(side="top")

    # title
    title_label = ttk.Label(
        master=window,
        text="Publish you Workflow into the QMENTA Platform. "
        "The workflow should already exist in your local environment.",
    )
    title_label.pack(padx=10, pady=10)

    # title
    title_label = ttk.Label(
        master=window, text="(*) indicates mandatory field."
    )
    title_label.pack(padx=10, pady=10)

    user_id_frame = ttk.Frame(master=window)
    user_id_label = ttk.Label(master=user_id_frame, text="Username*")
    user_id_label.pack(side="left")
    user_id_entry = ttk.Entry(master=user_id_frame)
    user_id_entry.pack(side="left", padx=10)
    user_id_frame.pack(pady=10)

    password_frame = ttk.Frame(master=window)
    password_label = ttk.Label(master=password_frame, text="Password*")
    password_label.pack(side="left")
    password_entry = ttk.Entry(master=password_frame, show="*")
    password_entry.pack(side="left", padx=10)
    password_frame.pack(pady=10)

    # Create a button to browse for a folder
    folder_tool_frame = ttk.Frame(master=window)
    folder_tool_label = ttk.Label(
        master=folder_tool_frame,
        text="Select folder where the tool is stored.*",
    )
    folder_tool_label.pack(side="left")
    folder_tool_button = ttk.Button(
        master=folder_tool_frame, text="Browse Folder", command=browse_folder
    )
    folder_tool_button.pack(side="left", padx=10)
    folder_label = ttk.Label(master=folder_tool_frame, textvariable=write_dir)
    folder_label.pack(side="bottom", pady=10)
    folder_tool_frame.pack(pady=10)

    # input tool ID field
    tool_id_frame = ttk.Frame(master=window)
    tool_id_label = ttk.Label(
        master=tool_id_frame, text="Specify the Workflow ID.*"
    )
    tool_id_label.pack(side="left")
    tool_id_entry = ttk.Entry(master=tool_id_frame)
    tool_id_entry.pack(side="left", padx=10)
    tool_id_frame.pack(pady=10)

    # input tool version field
    tool_version_frame = ttk.Frame(master=window)
    tool_version_label = ttk.Label(
        master=tool_version_frame, text="Specify the tool version.*"
    )
    tool_version_label.pack(side="left")
    tool_version_entry = ttk.Entry(master=tool_version_frame)
    tool_version_entry.pack(side="left", padx=10)
    tool_version_frame.pack(pady=10)

    tool_name_frame = ttk.Frame(master=window)
    tool_name_label = ttk.Label(
        master=tool_name_frame, text="Specify the tool name.*"
    )
    tool_name_label.pack(side="left")
    tool_name_entry = ttk.Entry(master=tool_name_frame)
    tool_name_entry.pack(side="left", padx=10)
    tool_name_frame.pack(pady=10)
    gui_content = {}

    def submit_form():
        gui_content.update(generate_output_object())
        gui_content.update(perform_selection_check(gui_content))
        if gui_content:
            window.destroy()

    action_frame = ttk.Frame(master=window)
    button = ttk.Button(
        master=action_frame,
        text="Publish in QMENTA Platform",
        command=submit_form,
    )
    button.pack(side="left")
    action_frame.pack(pady=10)

    def quit_tool_creator():
        window.destroy()
        print("Closing GUI.")
        exit()

    ttk.Button(
        master=action_frame,
        text="Quit",
        command=quit_tool_creator,
        bootstyle="danger",
    ).pack(side="left", padx=10)

    # run
    window.mainloop()
    return gui_content


def main():
    content_build = gui_tkinter()
    if not content_build:
        exit()
    os.chdir(dir_name)

    user = content_build["qmenta_user"]
    password = content_build["qmenta_password"]
    try:
        auth = platform.Auth.login(
            username=user,
            password=password,
            base_url="https://platform.qmenta.com",
            ask_for_2fa_input=False,
        )
    except Needs2FAError as needs2faerror:
        messagebox.showinfo(
            "Message",
            str(needs2faerror)
            + " Please check the terminal to add the code sent to your phone.",
        )
        auth = platform.Auth.login(
            username=user,
            password=password,
            base_url="https://platform.qmenta.com",
            code_2fa=input("Input 2-FA code:"),
        )

    # Get information from the advanced options file.
    advanced_options = "settings.json"
    raise_if_false(
        os.path.exists(advanced_options),
        "Settings do not exist! Run the local test to create it.",
    )
    with open(advanced_options, "r") as file:
        content_build["advanced_options"] = file.read()

    # Get information from the description file.
    with open("description.html") as fr:
        content_build["description"] = fr.read()

    # Get information from the model file.
    with open("model.flow") as fr:
        content_build["model"] = fr.read()

    content_build["results_configuration"] = {"tools": [], "screen": {}}
    if os.path.exists("results_configuration.json"):
        with open("results_configuration.json", "r") as file:
            results_config = json.load(file)
        # The screen value is expected as a string with escaped chars (dict)
        results_config["screen"] = json.dumps(results_config["screen"])
        content_build["results_configuration"] = json.dumps(
            results_config
        ).replace("{}", "")

    content_build.update(
        {
            "start_condition_code": "output={'OK': True, 'code': 1}",
            "deprecated": "0",
            "average_time": "1",
            "tags": "dev",
        }
    )

    # After creating the workflow, the ID of the workflow must be requested and added to the previous dictionary
    # otherwise it will keep creating new workflows on the platform creating conflicts.
    res = platform.post(
        auth, "a_administrator/upsert_analysis_tool", data=content_build
    )
    if res.json()["success"] == 1:
        print("Tool updated successfully!")
        print(
            "Tool name:",
            content_build["name"],
            "(",
            content_build["code"],
            ":",
            content_build["version"],
            ")",
        )
    else:
        print("ERROR setting the tool.")
        print(res.json())


if __name__ == "__main__":
    main()



import importlib
import inspect
import logging

from enum import Enum

from qmenta.sdk.context import NoFilesError


class AnalysisState(Enum):
    RUNNING = "running"
    COMPLETED = "completed"
    EXCEPTION = "exception"
    NO_FILES = "no_files"


class ExecClient(object):
    def __init__(self, analysis_id, comm, context, log_path):
        self.__analysis_id = analysis_id
        self.__comm = comm
        self.__context = context
        self.__log_path = log_path

    def __call__(self, user_script_path):
        """
        Run an analysis provided by the user

        Parameters
        ----------
        user_script_path : str
            Should look like 'path.to.module:object', where 'object' should be an importable analysis tool class
            or a function taking one 'context' argument.
        """

        logger = logging.getLogger(__name__)
        logger.info("Launching: {}".format(user_script_path))

        analysis_state = None
        try:
            # noinspection PyTypeChecker
            self.set_state(AnalysisState.RUNNING, log_path=self.__log_path)
            class_or_func = load_user_analysis(user_script_path)
            self.__context.fetch_analysis_data()

            # run the analysis
            # may allow classes to be used later
            assert inspect.isfunction(class_or_func)
            self.__context.set_progress(value=0, message="Running")
            class_or_func(self.__context)
        except NoFilesError:
            logger.warning("No files found")
            analysis_state = AnalysisState.NO_FILES
        except Exception as e:
            logger.exception(e)
            self.__context.set_progress(message="Error")
            analysis_state = AnalysisState.EXCEPTION
            raise e
        else:
            self.__context.set_progress(message="Completed", value=100)
            analysis_state = AnalysisState.COMPLETED
        finally:
            self.upload_log()
            # noinspection PyTypeChecker
            self.set_state(analysis_state)

    def set_state(self, state, **kwargs):
        """
        Set analysis state

        Parameters
        ----------
        state : AnalysisState
            One of AnalysisState.RUNNING, AnalysisState.COMPLETED, AnalysisState.EXCEPTION or AnalysisState.NO_FILES
        kwargs : dict
        """
        assert isinstance(
            state, AnalysisState
        ), "'state' is {}, not an 'AnalysisState'".format(state)
        data_to_send = {
            "analysis_id": self.__analysis_id,
            "state": state.value,
        }
        data_to_send.update(kwargs)
        logging.getLogger(__name__).info(
            "State = {state!r} for analysis {analysis_id}".format(
                **data_to_send
            )
        )
        self.__comm.send_request(
            "analysis_manager/set_analysis_state", data_to_send
        )

    def upload_log(self):

        data_to_send = {"analysis_id": self.__analysis_id}
        log_contents = ""
        try:
            with open(self.__log_path, "r") as log_file:
                log_contents = log_file.read()
        except IOError:
            log_contents = "Log file error ({})".format(self.__log_path)
        finally:
            # Send file in one single request
            # FIXME Implement chuncked upload
            self.__comm.send_files(
                "analysis_manager/upload_log",
                files={"file": ("exec.log", log_contents)},
                req_data=data_to_send,
            )


def load_user_analysis(user_script_path):
    module_name, class_or_func_name = user_script_path.split(":")
    user_module = importlib.import_module(module_name)
    class_or_func = getattr(user_module, class_or_func_name)
    return class_or_func



import os

from qmenta.sdk.local.context import LocalFile
from .modalities import Modality


class TestFileInput(LocalFile):
    __test__ = False

    def __init__(
        self,
        path: str,
        file_filter_condition_name: str,
        modality: Modality = Modality.none,
        tags=None,
        file_info=None,
        file_name_path: str = "",
        regex: str = ".*",
        mandatory: int = 1,
    ):
        super().__init__(
            os.path.split(path)[0],
            os.path.split(path)[1],
            modality,
            tags,
            file_info,
        )
        if file_info is None:
            file_info = {}
        if tags is None:
            tags = list()
        self.file_name_path = (
            os.path.basename(path) if not file_name_path else file_name_path
        )
        self.path = path
        self.name = path
        self.file_filter_condition_name = file_filter_condition_name
        self.modality = modality.value
        self.tags = tags
        self.file_info = file_info
        self.regex = regex
        self.mandatory = mandatory



import os
from enum import Enum, unique


@unique
class Modality(Enum):
    T1 = "T1"
    T2 = "T2"
    PD = "PD"
    fMRI = "fMRI"
    dMRI = "DWI"
    DTI = "DTI"
    HARDI = "HARDI"
    DSI = "DSI"
    ASL = "ASL"
    SCALAR = "SCALAR"
    T2_star = "T2-star"
    MRS = "MRS"
    SWI = "SWI"
    MRA = "MRA"
    CT = "CT"
    PET = "PET"
    EP = "EP"
    none = ""


class Tag(str):
    """Tag class to distinguish from regular expression in file filter.
    This class acts like a string"""

    def __init__(self, string_tag):
        self._iwd = os.getcwd()
        self._tag = string_tag

    def __str__(self):
        return self._tag

    def __repr__(self):
        return repr(self._tag)



import argparse
import logging
import os
import sys

from qmenta.sdk import __version__
from qmenta.sdk.executor import log_levels
from qmenta.sdk.local.client import LocalExecClient
from qmenta.sdk.local.context import LocalAnalysisContext
from qmenta.sdk.local.parse_settings import parse_tool_settings


def parse_args():
    argp = argparse.ArgumentParser(
        description="Local executor for debugging and testing purposes (SDK version {})".format(
            __version__
        )
    )
    argp.add_argument(
        "settings",
        help="JSON file that defines the settings of the tool to be run. Uses same syntax as in platform. "
        "See qmenta_sdk/examples/",
    )
    argp.add_argument(
        "input_settings",
        help="JSON file that defines the input values for the settings.",
    )
    argp.add_argument(
        "src_folder",
        help="Source folder where the inputs of the analysis are stored."
        'That is you "get_files" from this folder, that represents an input container.',
    )
    argp.add_argument(
        "dst_folder",
        help="Destination folder where the outputs of the analysis will be stored."
        'That is, when you "upload_file", the file is stored in this folder.',
    )
    argp.add_argument(
        "--res-folder",
        help="Folder where the different template files required by the analysis are stored.",
        default=None,
    )
    argp.add_argument(
        "--tool-path",
        required=True,
        help="Path to the Python tool (package.module:function) to be executed",
    )
    argp.add_argument("--log-path", help="Log file")
    argp.add_argument(
        "--log-level",
        choices=log_levels,
        default="debug",
        help="Logging level",
    )

    # version
    argp.add_argument("--version", action="version", version=__version__)

    return argp.parse_args()


def setup_logging(log_level, log_path=os.path.abspath("exec.log")):
    # Logging settings
    root_logger = logging.getLogger()
    formatter = logging.Formatter(
        "[%(asctime)s] %(name)s:%(lineno)d %(levelname)s :: %(message)s"
    )
    root_logger.setLevel(log_level)

    # File handler
    file_handler = logging.FileHandler(log_path)
    file_handler.setFormatter(formatter)
    root_logger.addHandler(file_handler)

    # Console handler
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(formatter)
    root_logger.addHandler(console_handler)

    return log_path


def main():
    arguments = parse_args()

    os.makedirs(arguments.dst_folder, exist_ok=True)

    log_path = (
        arguments.log_path
        if arguments.log_path is not None
        else os.path.join(arguments.dst_folder, "exec.log")
    )
    setup_logging(log_levels[arguments.log_level], log_path=log_path)

    # Parse settings
    settings = parse_tool_settings(
        arguments.settings, arguments.input_settings
    )

    # Instantiate client
    context = LocalAnalysisContext(
        settings,
        arguments.src_folder,
        arguments.dst_folder,
        arguments.res_folder,
    )
    client = LocalExecClient(context, log_path)

    logger = logging.getLogger(__name__)
    logger.info("Using SDK version (local emulation): {}".format(__version__))

    # Run it
    try:
        client(arguments.tool_path)
    except Exception:
        raise SystemExit(1)


if __name__ == "__main__":
    main()



import os

from qmenta import client  # noqa: ignore=E402

SCRIPT = "tool.py"
SETTINGS = "settings.json"
RES_CONF = "results_configuration.json"
DESCRIPTION = "description.html"
local_dockerfile = "Dockerfile"
local_requirements = "requirements.txt"
TEST_TOOL = "test_tool.py"

default_requirements = ["qmenta-sdk-lib"]
src_dir = os.path.join(os.path.dirname(__file__), "templates_tool_maker")


def raise_if_false(evaluation, message):
    if not evaluation:
        raise AssertionError(message)


def build_local_dockerfile() -> None:
    """
    Building the Dockerfile in the local folder to run locally
    """
    # Here we are in the local/ folder, we switch to the test/ folder
    if not os.path.exists(local_dockerfile):
        with open(local_dockerfile, "w") as w1:
            with open(os.path.join(src_dir, "Dockerfile_schema")) as r1:
                w1.write(r1.read())


def build_local_requirements() -> None:
    """
    Preparing the python requirements file in the local folder to run locally

    """
    if not os.path.exists(local_requirements):
        with open(local_requirements, "w") as f1:
            f1.writelines(
                [
                    "# Python requirements for tool:\n",
                    "# Ex: matplotlib==3.4.2\n",
                ]
                + default_requirements
            )


def build_script(code: str) -> None:
    """

    Writing the tool class with the tool_inputs(), run(), and tool_outputs() examples

    Parameters
    ----------
    code: str
        Tool ID provided by the GUI

    """
    with open(os.path.join(src_dir, "tool_schema")) as r1:
        with open(SCRIPT, "w") as w1:
            w1.write(
                r1.read().replace("TOOL_CLASS", convert_to_camel_case(code))
            )


def build_description() -> None:
    """
    Write an empty file, to be filled by the developer
    """
    with open(os.path.join(src_dir, "description_schema")) as r1:
        with open(DESCRIPTION, "w") as w1:
            w1.write(r1.read())


def convert_to_camel_case(code: str) -> str:
    """
    Utils function to convert the ID of the tool to camel case convention string for the class of the tool and the test.

    Parameters
    ----------
    code : str
        Tool ID provided by the GUI

    Returns
    -------
    str
    """
    return "".join([t.capitalize() for t in code.split("_")])


def build_test(code: str, tool_folder: str, version: str) -> None:
    """
    Create internal structure for the local folder, test and sample data

    Parameters
    ----------
    code : str
        Tool ID provided by the GUI
    tool_folder : str
        Tool folder provided by the GUI
    version : str
        Tool version provided by the GUI

    """
    os.makedirs("test/", exist_ok=True)
    os.makedirs("test/sample_data", exist_ok=True)
    os.chdir("test")
    with open(os.path.join(src_dir, "test_tool_schema")) as r1:
        with open(TEST_TOOL, "w") as w1:
            w1.write(
                r1.read()
                .replace("TOOL_CLASS", convert_to_camel_case(code))
                .replace("TOOL_ID", code)
                .replace("TOOL_FOLDER", os.path.join(tool_folder, code))
                .replace("VERSION", version)
            )



import sys
from argparse import ArgumentParser

entry_point_contents = """#!/bin/bash -eu
# Add your configuration here:
# ...

# Tool start:
exec {executable} -m qmenta.sdk.executor "$@"
"""


def make_entrypoint():

    parser = ArgumentParser()
    parser.add_argument("target")
    options = parser.parse_args()

    with open(options.target, "w") as fp:
        fp.write(entry_point_contents.format(executable=sys.executable))


if __name__ == "__main__":
    make_entrypoint()



import logging

from qmenta.sdk.client import ExecClient


class LocalExecClient(ExecClient):  # Override main execution client
    def __init__(self, context, log_path):
        super(LocalExecClient, self).__init__(
            analysis_id="local", comm=None, context=context, log_path=log_path
        )

    def set_state(self, state, **kwargs):
        """
        Set analysis state

        Parameters
        ----------
        state : AnalysisState
            One of AnalysisState.RUNNING, AnalysisState.COMPLETED, AnalysisState.EXCEPTION or AnalysisState.NO_FILES
        kwargs : dict
        """
        logging.getLogger(__name__).info(
            "State = {state!r} for local execution".format(state=state.value)
        )

    def upload_log(self):
        pass



import shutil
import tempfile


class TemporaryDirectory:
    """
    Context manager which creates a temporary directory and removes it automatically.
    """

    def __init__(self):
        self.name = None

    def __enter__(self):
        self.name = tempfile.mkdtemp()
        return self.name

    def __exit__(self, exc_type, exc_val, exc_tb):
        shutil.rmtree(self.name)



import logging
import os
from numbers import Number
from types import SimpleNamespace
from typing import List, Tuple

from qmenta.sdk.local.context import LocalAnalysisContext

Option_type = Tuple[object, str]
Options_type = List[Option_type]


def check_options(options, id_):
    assert all(
        [isinstance(o[1], str) for o in options]
    ), f"Title of setting {id_} must by of type 'str'"
    assert len([o[0] for o in options]) == len(
        list(set([o[0] for o in options]))
    ), f"{id_} must unique options."


class Input:
    def __init__(self, id_: str, input_type: str):
        self.id = id_
        self.type = input_type


class CheckBox(Input):
    def __init__(self, id_: str, title: str, default: int = 1):
        super().__init__(id_, "checkbox")
        self.id = id_
        self.title = title
        self.default = default

    def get_input(self, context):
        settings = context.get_settings()
        return bool(settings.get(self.id, self.default))


class String(Input):
    def __init__(self, id_: str, title: str, default: str):
        super().__init__(id_, "string")
        self.id = id_
        self.title = title
        self.default = default

    def get_input(self, context):
        settings = context.get_settings()
        return str(settings.get(self.id, self.default))


class Decimal(Input):
    def __init__(
        self,
        id_: str,
        title: str,
        default: Number,
        _min: Number = -1e9,
        _max: Number = 1e9,
    ):
        super().__init__(id_, "decimal")
        self.id = id_
        self.title = title
        self.default = default
        self.min = _min
        self.max = _max

    def get_input(self, context):
        settings = context.get_settings()
        return float(settings.get(self.id, self.default))


class Integer(Input):
    def __init__(
        self,
        id_: str,
        title: str,
        default: Number,
        _min: Number = 0,
        _max: Number = 1e9,
    ):
        super().__init__(id_, "integer")
        self.id = id_
        self.title = title
        self.default = default
        self.min = _min
        self.max = _max

    def get_input(self, context):
        settings = context.get_settings()
        return int(settings.get(self.id, self.default))


class SingleChoice(Input):
    def __init__(
        self, id_: str, options: Options_type, default: object, title: str
    ):
        super().__init__(id_, "single_choice")
        self.id = id_
        self.title = title
        self.options = options
        self.default = default
        check_options(options, id_)

    def get_input(self, context):
        return context.get_settings().get(self.id, self.default)


class MultipleChoice(Input):
    def __init__(
        self,
        id_: str,
        options: Options_type,
        default: List[object],
        title: str,
    ):
        super().__init__(id_, "multiple_choice")
        self.id = id_
        self.title = title
        self.options = options
        self.default = default
        check_options(options, id_)

    def get_input(self, context):
        return context.get_settings().get(self.id, self.default)


class ContainerHandler(Input):
    def __init__(
        self,
        id_: str,
        file_filter: str,
        title: str = "",
        info: str = "",
        anchor: int = 1,
        batch: int = 1,
        mandatory: int = 1,
        tool_codes_set_as_input=None,
    ):
        super().__init__(id_, "container")

        if tool_codes_set_as_input is None:
            tool_codes_set_as_input = list()
        self.file_filter = file_filter
        self.title = title
        if info:
            self.info = info
        self.in_filter = ["mri_brain_data"]
        self.out_filter = tool_codes_set_as_input
        self.anchor = anchor
        self.batch = batch
        self.mandatory = mandatory

    def get_input(self, context):
        """

        Parameters
        ----------
        context

        Returns
        -------

        """

        logger = logging.getLogger(__name__)
        analysis_dir = os.environ.get("MINTEXE_PATH", "/qmenta")

        # Download files and point to its filename
        fhandlers = SimpleNamespace()

        if isinstance(context, LocalAnalysisContext):
            for k, v in context._LocalAnalysisContext__settings.items():
                if isinstance(v, dict) and "files" in v:
                    context._LocalAnalysisContext__settings[k] = (
                        context._LocalAnalysisContext__settings[k]["files"]
                    )

        filter_names = list()
        for filter_name in self.file_filter.split(
            "["
        ):  # captures the end of file filter name
            filter_names.append(
                filter_name.split(" ")[-1].replace("(", "")
            )  # gets the last string (file filter name)

        for file_filter_condition_name in filter_names[
            :-1
        ]:  # the last element is not the file filter name.
            logger.info(f"Getting files from : {file_filter_condition_name}")
            fhandlers_ = context.get_files(
                self.id, file_filter_condition_name=file_filter_condition_name
            )
            for fh in fhandlers_:
                logger.info(f"Getting handlers from {fh}.")
                logger.info(f"Getting name handler {fh.name}.")
                if "/" in fh.name:
                    folder_to_save_file = ""
                else:
                    folder_to_save_file = os.path.dirname(fh.name)
                if fh.name.endswith(".zip"):
                    folder_to_save_file = os.path.basename(fh.name).replace(
                        ".zip", ""
                    )
                    if "/" in fh.name:
                        folder_to_save_file = fh.name.replace(".zip", "")

                fpath = os.path.join(
                    analysis_dir, "input_folder", self.id, folder_to_save_file
                )

                # This takes into account the path of the file in the input container in the platform
                logger.info(f"Saving to path: {os.path.dirname(fpath)}.")

                # in local testing, we need to create the directory
                os.makedirs(os.path.dirname(fpath), exist_ok=True)

                fh.file_path = fh.download(fpath)

            fhandlers.__setattr__(file_filter_condition_name, fhandlers_)

        return fhandlers


class SubjectHandler(Input):
    def __init__(
        self,
        id_: str,
        file_filter: str,
        title: str = "",
        info: str = "",
        anchor: int = 1,
        batch: int = 1,
        mandatory: int = 1,
        tool_codes_set_as_input=None,
    ):
        super().__init__(id_, "subject")

        if tool_codes_set_as_input is None:
            tool_codes_set_as_input = list()
        self.file_filter = file_filter
        self.title = title
        self.info = info
        self.in_filter = ["mri_brain_data"]
        self.out_filter = tool_codes_set_as_input
        self.anchor = anchor
        self.batch = batch
        self.mandatory = mandatory

    def get_input(self, context):
        pass  # pragma: no cover


class Heading:
    def __init__(self, content: str):
        self.type = "heading"
        self.content = content

    def get_input(self, context):
        pass  # pragma: no cover


class Line:
    def __init__(self):
        self.type = "line"

    def get_input(self, context):
        pass  # pragma: no cover


class IndentText:
    def __init__(self, content: str):
        self.type = "indent"
        self.content = content

    def get_input(self, context):
        pass  # pragma: no cover


class InfoText:
    def __init__(self, content: str):
        self.type = "info"
        self.content = content

    def get_input(self, context):
        pass  # pragma: no cover



from typing import List, Union

from .modalities import Modality, Tag

Property_type = Union[Modality, List[Modality], List[str], str, int]


def make_tag(p: Union[List[Tag]]) -> str:
    """
    Prepends the tag t' to each string provided as argument or in the list.
    If the tag starts with "!" then the prefix is "!t'" instead.

    Parameters
    ----------
    p: tag or list of tags.

    Returns
    -------
    str
    """
    return ",".join(
        [
            f"t'{tag}'" if not tag.startswith("!") else f"!t'{tag[1:]}'"
            for tag in p
        ]
    )


def make_modality(p: Union[List[Modality], Modality]) -> str:
    return (
        "|".join([f"m'{mod.value}'" for mod in p])
        if (isinstance(p, list) and len(p) > 1)
        else f"m'{p.value}'"
    )


def make_regex(p: str) -> str:
    return f"r'{p}'"


def process_element(seq: Property_type) -> str:
    if isinstance(seq, list) and len(seq) and isinstance(seq[0], Tag):
        return make_tag(seq)
    elif isinstance(seq, Tag):
        return make_tag([seq])
    elif (isinstance(seq, Modality) and seq.value != "None") or (
        isinstance(seq, list) and len(seq) > 1 and isinstance(seq[0], Modality)
    ):
        return make_modality(seq)
    elif isinstance(seq, list) and len(seq) and isinstance(seq[0], Modality):
        return make_modality(seq[0])
    elif isinstance(seq, list) and len(seq) and isinstance(seq[0], list):
        list_output = list()
        for s in seq:
            list_output.append([process_element(i) for i in s])
        nlist = list()
        for el in list_output:
            el2 = "(" + ",".join(el) + ")"
            nlist.append(el2)
        return "|".join(nlist)
    elif isinstance(seq, str):
        return make_regex(seq)


def parse_property(filters: List[Property_type]) -> str:
    parsed = list()
    if not filters:
        return make_regex(".*")
    for _filter in filters:
        pf = process_element(_filter)
        (
            parsed.append("(" + pf + ")")
            if ("," in pf or "|" in pf) and not isinstance(_filter, str)
            else parsed.append(pf)
        )
    return (
        "(" + ",".join(parsed) + ")" if len(filters) > 1 else ",".join(parsed)
    )


def parse_file_filter_property(conditions: List) -> str:
    parsed = [parse_property(condition[0]) for condition in conditions]
    return parsed[0] if len(parsed) == 1 else ("(" + "|".join(parsed) + ")")


class ConditionForFilter(object):
    def __init__(
        self,
        file_filter_condition_name: str,
        condition: List[tuple],
        min_files: int = 1,
        max_files: Union[int, str] = 1,
    ):
        self.id = file_filter_condition_name
        self.condition = condition
        self.property = parse_file_filter_property(condition)
        assert min_files <= max_files if max_files != "*" else True
        self.number_files = f"[{min_files},{max_files}]"

    def __str__(self):
        return f"{self.number_files}({self.property})"



import sys
from argparse import ArgumentParser

entry_point_contents = """#!/bin/bash -exu
export PYTHONPATH="${{PYTHONPATH:+$PYTHONPATH:}}{tool_path}"

# Add your configuration here:
# ...

# Source the export_paths.txt file if exist to load the system paths.
# This file is created in the ci when performing a multi-level build of the docker images

if [ -f "/root/export_paths.txt" ]; then
source /root/export_paths.txt
fi


# Tool start:
exec {executable} -m qmenta.sdk.executor "$@"
"""


def make_entrypoint():
    parser = ArgumentParser()
    parser.add_argument("target")
    parser.add_argument("tool_paths", nargs="+")
    options = parser.parse_args()

    with open(options.target, "w") as fp:
        fp.write(
            entry_point_contents.format(
                executable=sys.executable,
                tool_path=":".join(options.tool_paths),
            )
        )


if __name__ == "__main__":
    make_entrypoint()
